import unittest

import numpy as np

from abcpy.output import Journal


class JournalTests(unittest.TestCase):
    # def test_add_parameters(self):
    #     params1 = np.zeros((2,4))
    #     params2 = np.ones((2,4))
    #
    #     # test whether production mode only stores the last set of parameters
    #     journal_prod = Journal(0)
    #     journal_prod.add_parameters(params1)
    #     journal_prod.add_parameters(params2)
    #     self.assertEqual(len(journal_prod.parameters), 1)
    #     np.testing.assert_equal(journal_prod.parameters[0], params2)
    #
    #     # test whether reconstruction mode stores all parameter sets
    #     journal_recon = Journal(1)
    #     journal_recon.add_parameters(params1)
    #     journal_recon.add_parameters(params2)
    #     self.assertEqual(len(journal_recon.parameters), 2)
    #     np.testing.assert_equal(journal_recon.parameters[0], params1)
    #     np.testing.assert_equal(journal_recon.parameters[1], params2)

    def test_add_weights(self):
        weights1 = np.zeros((2, 4))
        weights2 = np.ones((2, 4))

        # test whether production mode only stores the last set of parameters
        journal_prod = Journal(0)
        journal_prod.add_weights(weights1)
        journal_prod.add_weights(weights2)
        self.assertEqual(len(journal_prod.weights), 1)
        np.testing.assert_equal(journal_prod.weights[0], weights2)

        # test whether reconstruction mode stores all parameter sets
        journal_recon = Journal(1)
        journal_recon.add_weights(weights1)
        journal_recon.add_weights(weights2)
        self.assertEqual(len(journal_recon.weights), 2)
        np.testing.assert_equal(journal_recon.weights[0], weights1)
        np.testing.assert_equal(journal_recon.weights[1], weights2)

    def test_add_opt_values(self):
        opt_values1 = np.zeros((2, 4))
        opt_values2 = np.ones((2, 4))

        # test whether production mode only stores the last set of parameters
        journal_prod = Journal(0)
        journal_prod.add_opt_values(opt_values1)
        journal_prod.add_opt_values(opt_values2)
        self.assertEqual(len(journal_prod.opt_values), 1)
        np.testing.assert_equal(journal_prod.opt_values[0], opt_values2)

        # test whether reconstruction mode stores all parameter sets
        journal_recon = Journal(1)
        journal_recon.add_opt_values(opt_values1)
        journal_recon.add_opt_values(opt_values2)
        self.assertEqual(len(journal_recon.opt_values), 2)
        np.testing.assert_equal(journal_recon.opt_values[0], opt_values1)
        np.testing.assert_equal(journal_recon.opt_values[1], opt_values2)

    def test_load_and_save(self):
        params1 = np.zeros((2, 4))
        weights1 = np.zeros((2, 4))

        journal = Journal(0)
        # journal.add_parameters(params1)
        journal.add_weights(weights1)
        journal.save('journal_tests_testfile.pkl')

        new_journal = Journal.fromFile('journal_tests_testfile.pkl')
        # np.testing.assert_equal(journal.parameters, new_journal.parameters)
        np.testing.assert_equal(journal.weights, new_journal.weights)

    def test_ESS(self):
        weights_identical = np.ones((100, 1))
        weights = np.arange(100).reshape(-1, 1)
        journal = Journal(1)
        journal.add_weights(weights_identical)
        journal.add_weights(weights)
        journal.add_ESS_estimate(weights=weights_identical)
        journal.add_ESS_estimate(weights=weights)
        self.assertEqual(len(journal.ESS), 2)
        self.assertAlmostEqual(journal.get_ESS_estimates(), 74.62311557788945)
        self.assertAlmostEqual(journal.get_ESS_estimates(0), 100)

    def test_plot_ESS(self):
        weights_identical = np.ones((100, 1))
        weights_1 = np.arange(100).reshape(-1, 1)
        weights_2 = np.arange(100, 200).reshape(-1, 1)
        journal = Journal(1)
        journal.add_weights(weights_identical)
        journal.add_ESS_estimate(weights=weights_identical)
        journal.add_weights(weights_1)
        journal.add_ESS_estimate(weights=weights_1)
        journal.add_weights(weights_2)
        journal.add_ESS_estimate(weights=weights_2)
        journal.plot_ESS()
        journal_2 = Journal(0)
        self.assertRaises(RuntimeError, journal_2.plot_ESS)

    def test_plot_wass_dist(self):
        rng = np.random.RandomState(1)
        weights_identical = np.ones((100, 1))
        params_0 = rng.randn(100).reshape(-1, 1)
        weights_1 = np.arange(100)
        params_1 = rng.randn(100).reshape(-1, 1, 1)
        weights_2 = np.arange(100, 200)
        params_2 = rng.randn(100).reshape(-1, 1)
        weights_3 = np.arange(200, 300)
        params_3 = rng.randn(100).reshape(-1, 1)
        weights_4 = np.arange(300, 400)
        params_4 = rng.randn(100).reshape(-1, 1)
        journal = Journal(1)
        journal.add_weights(weights_identical)
        journal.add_accepted_parameters(params_0)
        journal.add_weights(weights_1)
        journal.add_accepted_parameters(params_1)
        journal.add_weights(weights_2)
        journal.add_accepted_parameters(params_2)
        journal.add_weights(weights_3)
        journal.add_accepted_parameters(params_3)
        journal.add_weights(weights_4)
        journal.add_accepted_parameters(params_4)
        fig, ax, wass_dist_lists = journal.Wass_convergence_plot()
        self.assertAlmostEqual(wass_dist_lists[0], 0.22829193592175878)
        # check the Errors
        journal_2 = Journal(0)
        self.assertRaises(RuntimeError, journal_2.Wass_convergence_plot)
        journal_3 = Journal(1)
        journal_3.add_weights(weights_identical)
        self.assertRaises(RuntimeError, journal_3.Wass_convergence_plot)
        journal_4 = Journal(1)
        journal_4.add_accepted_parameters(np.array([np.array([1]), np.array([1, 2])], dtype="object"))
        print(len(journal_4.accepted_parameters))
        self.assertRaises(RuntimeError, journal_4.Wass_convergence_plot)

    def test_plot_post_distr(self):
        rng = np.random.RandomState(1)
        weights_identical = np.ones((100, 1))
        params = rng.randn(100, 2, 1, 1)
        weights = np.arange(100).reshape(-1, 1)
        journal = Journal(1)
        journal.add_user_parameters([("par1", params[:, 0]), ("par2", params[:, 1])])
        journal.add_user_parameters([("par1", params[:, 0]), ("par2", params[:, 1])])
        journal.add_weights(weights=weights_identical)
        journal.add_weights(weights=weights)
        journal.plot_posterior_distr(single_marginals_only=True, iteration=0)
        journal.plot_posterior_distr(true_parameter_values=[0.5, 0.3], show_samples=True)
        journal.plot_posterior_distr(double_marginals_only=True, show_samples=True,
                                     true_parameter_values=[0.5, 0.3])
        journal.plot_posterior_distr(contour_levels=10, ranges_parameters={"par1": [-1, 1]},
                                     parameters_to_show=["par1"])

        with self.assertRaises(KeyError):
            journal.plot_posterior_distr(parameters_to_show=["par3"])
        with self.assertRaises(RuntimeError):
            journal.plot_posterior_distr(single_marginals_only=True, double_marginals_only=True)
        with self.assertRaises(RuntimeError):
            journal.plot_posterior_distr(parameters_to_show=["par1"], double_marginals_only=True)
        with self.assertRaises(RuntimeError):
            journal.plot_posterior_distr(parameters_to_show=["par1"], true_parameter_values=[0.5, 0.3])
        with self.assertRaises(TypeError):
            journal.plot_posterior_distr(ranges_parameters={"par1": [-1]})
        with self.assertRaises(TypeError):
            journal.plot_posterior_distr(ranges_parameters={"par1": np.zeros(1)})

    def test_traceplot(self):
        rng = np.random.RandomState(1)
        weights_identical = np.ones((100, 1))
        params = rng.randn(100).reshape(-1, 1)
        journal = Journal(1)
        journal.add_weights(weights_identical)
        journal.add_accepted_parameters(params)
        journal.add_user_parameters([("mu", params[:, 0])])
        self.assertRaises(RuntimeError, journal.traceplot)  # as it does not have "acceptance_rates" in configuration
        journal.configuration["acceptance_rates"] = [0.3]
        with self.assertRaises(KeyError):
            journal.traceplot(parameters_to_show=["sigma"])
        # now try correctly:
        fig, ax = journal.traceplot()


if __name__ == '__main__':
    unittest.main()



import pickle
import unittest

import cloudpickle
import numpy as np

'''We use pickle in our MPI backend to send a method from the master to the workers. The object with which this method is associated cotains the backend as an attribute, while the backend itself contains the data on which the workers should work. Pickling the method results in pickling the backend, which results in the whole data being pickled and sent, which is undersirable.

In pickle, the "getstate" method can be specified. When an object cotaining a "getstate" method is pickled, only the attributes specified within that method are pickled.

This test checks whether everything is working correctly with cloudpickle.
'''


class ToBePickled:
    def __init__(self):
        self.included = 5
        self.notIncluded = np.zeros(10 ** 5)

    def __getstate__(self):
        """Method that tells cloudpickle which attributes should be pickled
        Returns
        -------
        state
            all the attributes that should be pickled
        """
        state = self.__dict__.copy()
        del state['notIncluded']
        return state


class PickleTests(unittest.TestCase):
    def test_exclusion(self):
        """Tests whether after pickling and unpickling the object, the attribute which should not be included exists"""
        pickled_object = cloudpickle.dumps(ToBePickled(), pickle.HIGHEST_PROTOCOL)
        unpickled_object = cloudpickle.loads(pickled_object)
        self.assertTrue(not (hasattr(pickled_object, 'notIncluded')))


if __name__ == '__main__':
    unittest.main()



try:
    import torch
except ImportError:
    has_torch = False
else:
    has_torch = True

import logging

import numpy as np


def dist2(x, y):
    """Compute the square of the Euclidean distance between 2 arrays of same length"""
    return np.dot(x - y, x - y)


def compute_similarity_matrix(target, quantile=0.1, return_pairwise_distances=False):
    """Compute the similarity matrix between some values given a given quantile of the Euclidean distances.

    If return_pairwise_distances is True, it also returns a matrix with the pairwise distances with every distance."""

    logger = logging.getLogger("Compute_similarity_matrix")

    n_samples = target.shape[0]

    pairwise_distances = np.zeros([n_samples] * 2)

    for i in range(n_samples):
        for j in range(n_samples):
            pairwise_distances[i, j] = dist2(target[i], target[j])

    q = np.quantile(pairwise_distances[~np.eye(n_samples, dtype=bool)].reshape(-1), quantile)

    similarity_set = pairwise_distances < q

    logger.info("Fraction of similar pairs (epurated by self-similarity): {}".format(
        (np.sum(similarity_set) - n_samples) / n_samples ** 2))

    if (np.sum(similarity_set) - n_samples) / n_samples ** 2 == 0:
        raise RuntimeError("The chosen quantile is too small, as there are no similar samples according to the "
                           "corresponding threshold.\nPlease increase the quantile.")

    return (similarity_set, pairwise_distances) if return_pairwise_distances else similarity_set


def save_net(path, net):
    """Function to save the Pytorch state_dict of a network to a file."""
    torch.save(net.state_dict(), path)


def load_net(path, network_class, *network_args, **network_kwargs):
    """Function to load a network from a Pytorch state_dict, given the corresponding network_class."""
    net = network_class(*network_args, **network_kwargs)
    net.load_state_dict(torch.load(path))
    return net.eval()  # call the network to eval model. Needed with batch normalization and dropout layers.



# noinspection PyInterpreter
import logging
import pickle
import time

import cloudpickle
import numpy as np
from mpi4py import MPI

import abcpy.backends.mpimanager
from abcpy.backends import BDS, PDS, Backend, NestedParallelizationController


class NestedParallelizationControllerMPI(NestedParallelizationController):
    def __init__(self, mpi_comm):
        self.logger = logging.getLogger(__name__)
        self.logger.info("#### Initialize NPC ####")
        self.loop_workers = True
        self.mpi_comm = mpi_comm
        self.nested_func = "NoFunction"
        self.func_args = ()
        self.func_kwargs = {}
        self.result = None
        if self.mpi_comm.Get_rank() != 0:
            self.nested_execution()

    def communicator(self):
        return self.mpi_comm

    def nested_execution(self):
        rank = self.mpi_comm.Get_rank()
        self.logger.debug("Starting nested loop on rank {}".format(rank))
        while self.loop_workers:
            self.mpi_comm.barrier()
            self.loop_workers = self.mpi_comm.bcast(self.loop_workers, root=0)
            if self.loop_workers == False:
                return
            func_p = None
            func_args_p = None
            func_kwargs_p = None
            if self.mpi_comm.Get_rank() == 0:
                self.logger.debug("Start pickling func on rank {}".format(rank))
                func_p = cloudpickle.dumps(self.nested_func, pickle.HIGHEST_PROTOCOL)
                func_args_p = cloudpickle.dumps(self.func_args, pickle.HIGHEST_PROTOCOL)
                func_kwargs_p = cloudpickle.dumps(self.func_kwargs, pickle.HIGHEST_PROTOCOL)

            self.logger.debug("Broadcasting function {} on rank {}".format(self.nested_func, rank))
            func_p = self.mpi_comm.bcast(func_p, root=0)
            func_args_p = self.mpi_comm.bcast(func_args_p, root=0)
            func_kwargs_p = self.mpi_comm.bcast(func_kwargs_p, root=0)
            self.nested_func = cloudpickle.loads(func_p)
            self.func_args = cloudpickle.loads(func_args_p)
            self.func_kwargs = cloudpickle.loads(func_kwargs_p)

            func = self.nested_func
            self.logger.debug("Starting map function {} on rank {}".format(func.__name__, self.mpi_comm.Get_rank()))
            self.func_kwargs['mpi_comm'] = self.mpi_comm
            self.mpi_comm.barrier()
            self.result = func(*self.func_args, **self.func_kwargs)
            self.logger.debug("Ending map function on rank {}".format(self.mpi_comm.Get_rank()))
            self.mpi_comm.barrier()
            if self.mpi_comm.Get_rank() == 0:
                return
        self.loop_workers = True
        self.logger.debug("Ending nested loop on rank {}".format(self.mpi_comm.Get_rank()))

    def run_nested(self, func, *args, **kwargs):
        self.logger.debug("Executing nested function {}.".format(func.__name__))
        self.nested_func = func
        self.func_args = args
        self.func_kwargs = kwargs
        self.nested_execution()
        self.logger.debug("Return from nested execution of master rank")
        self.nested_func = None
        self.func_args = ()
        self.func_kwargs = {}
        self.logger.info(self.result)
        return self.result

    def __del__(self):
        rank = self.mpi_comm.Get_rank()
        self.logger.debug("Stopping npc on rank {}".format(rank))
        self.loop_workers = False
        if rank == 0:
            self.mpi_comm.barrier()
            self.loop_workers = self.mpi_comm.bcast(self.loop_workers, root=0)
        self.logger.debug(">>>>>>>> NPC stopped on rank {}".format(rank))


class BackendMPIScheduler(Backend):
    """Defines the behavior of the scheduler process

    This class defines the behavior of the scheduler process (The one
    with rank==0) in MPI.

    """

    # Define some operation codes to make it more readable
    OP_PARALLELIZE, OP_MAP, OP_COLLECT, OP_BROADCAST, OP_DELETEPDS, OP_DELETEBDS, OP_FINISH = [1, 2, 3, 4, 5, 6, 7]
    finalized = False

    def __init__(self, chunk_size=1):
        """
        Parameters
        ----------
        chunk_size: Integer
            size of one block of data to be sent to free
            execution teams
       """

        # Initialize the current_pds_id and bds_id
        self.__current_pds_id = 0
        self.__current_bds_id = 0

        # Initialize a BDS store for both scheduler & team.
        self.bds_store = {}
        self.pds_store = {}

        # Initialize a store for the pds data that
        # .. hasn't been sent to the teams yet
        self.pds_pending_store = {}

        self.chunk_size = chunk_size

    def __command_teams(self, command, data):
        """Tell teams to enter relevant execution block
        This method handles the sending of the command to the teams
        telling them what operation to perform next.

        Parameters
        ----------
        command: operation code of OP_xxx
            One of the operation codes defined in the class definition as OP_xxx
            which tell the teams what operation they're performing.
        data:  tuple
            Any of the data required for the operation which needs to be bundled
            in the data packet sent.
        """

        if command == self.OP_PARALLELIZE:
            # In parallelize we receive data as (pds_id)
            data_packet = (command, data[0])

        elif command == self.OP_MAP:
            # In map we receive data as (pds_id,pds_id_new,func)
            # Use cloudpickle to dump the function into a string.
            function_packed = cloudpickle.dumps(data[2], pickle.HIGHEST_PROTOCOL)
            data_packet = (command, data[0], data[1], function_packed)

        elif command == self.OP_BROADCAST:
            data_packet = (command, data[0])

        elif command == self.OP_COLLECT:
            # In collect we receive data as (pds_id)
            data_packet = (command, data[0])

        elif command == self.OP_DELETEPDS or command == self.OP_DELETEBDS:
            # In deletepds we receive data as (pds_id) or bds_id
            data_packet = (command, data[0])

        elif command == self.OP_FINISH:
            data_packet = (command,)

        _ = self.mpimanager.get_scheduler_communicator().bcast(data_packet, root=0)

    def __generate_new_pds_id(self):
        """
        This method generates a new pds_id to associate a PDS with it's remote counterpart
        that teams use to store & index data based on the pds_id they receive

        Returns
        -------
        Returns a unique integer id.

        """

        self.__current_pds_id += 1
        return self.__current_pds_id

    def __generate_new_bds_id(self):
        """
        This method generates a new bds_id to associate a BDS with it's remote counterpart
        that teams use to store & index data based on the bds_id they receive

        Returns
        -------
        Returns a unique integer id.

        """

        self.__current_bds_id += 1
        return self.__current_bds_id

    def parallelize(self, python_list):
        """
        This method distributes the list on the available teams and returns a
        reference object.

        The list is split into number of teams many parts as a numpy array.
        Each part is sent to a separate team node using the MPI scatter.

        scheduler: python_list is the real data that is to be split up

        Parameters
        ----------
        list: Python list
            the list that should get distributed on the leader nodes of the teams

        Returns
        -------
        PDSMPI class (parallel data set)
            A reference object that represents the parallelized list
        """

        # Tell the teams to enter parallelize()
        pds_id = self.__generate_new_pds_id()
        self.__command_teams(self.OP_PARALLELIZE, (pds_id,))

        # Don't send any data. Just keep it as a queue we're going to pop.
        self.pds_store[pds_id] = list(python_list)

        pds = PDSMPI([], pds_id, self)

        return pds

    def orchestrate_map(self, pds_id):
        """Orchestrates the teams to perform a map function
        
        This works by keeping track of the teams who haven't finished executing,
        waiting for them to request the next chunk of data when they are free,
        responding to them with the data and then sending them a Sentinel
        signalling that they can exit.
        """
        is_map_done = [True if i in self.mpimanager.get_scheduler_node_ranks() else False for i in
                       range(self.mpimanager.get_scheduler_size())]
        status = MPI.Status()

        # Copy it to the pending. This is so when scheduler accesses
        # the PDS data it's not empty.
        self.pds_pending_store[pds_id] = list(self.pds_store[pds_id])

        # While we have some ranks that haven't finished
        while sum(is_map_done) < self.mpimanager.get_scheduler_size():
            # Wait for a request from anyone
            data_request = self.mpimanager.get_scheduler_communicator().recv(
                source=MPI.ANY_SOURCE,
                tag=MPI.ANY_TAG,
                status=status,
            )
            request_from_rank = status.source

            if data_request != pds_id:
                print("Ignoring stale PDS data request from",
                      request_from_rank, ":", data_request, "/", pds_id)
                continue

            # Pointer so we don't have to keep doing dict lookups
            current_pds_items = self.pds_pending_store[pds_id]
            num_current_pds_items = len(current_pds_items)

            # Everyone's already exhausted all the data.
            # Send a sentinel and mark the node as finished
            if num_current_pds_items == 0:
                self.mpimanager.get_scheduler_communicator().send(None, dest=request_from_rank, tag=pds_id)
                is_map_done[request_from_rank] = True
            else:
                # Create the chunk of data to send. Pop off items and tag them with an id.
                # so we can sort them later
                chunk_to_send = []
                for i in range(self.chunk_size):
                    chunk_to_send += [(num_current_pds_items - i, current_pds_items.pop())]
                    self.mpimanager.get_scheduler_communicator().send(chunk_to_send, dest=request_from_rank, tag=pds_id)

    def map(self, func, pds):
        """
        A distributed implementation of map that works on parallel data sets (PDS).

        On every element of pds the function func is called.

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDS class
            A parallel data set to which func should be applied

        Returns
        -------
        PDSMPI class
            a new parallel data set that contains the result of the map
        """

        # Tell the teams to enter the map() with the current pds_id & func.
        # Get pds_id of dataset we want to operate on
        pds_id = pds.pds_id

        # Generate a new pds_id to be used by the teams for the resultant PDS
        pds_id_new = self.__generate_new_pds_id()

        data = (pds_id, pds_id_new, func)
        self.__command_teams(self.OP_MAP, data)
        self.orchestrate_map(pds_id)

        pds_res = PDSMPI([], pds_id_new, self)

        return pds_res

    def collect(self, pds):
        """
        Gather the pds from all the teams,
            send it to the scheduler and return it as a standard Python list.

        Parameters
        ----------
        pds: PDS class
            a parallel data set

        Returns
        -------
        Python list
            all elements of pds as a list
        """

        # Tell the teams to enter collect with the pds's pds_id
        self.__command_teams(self.OP_COLLECT, (pds.pds_id,))

        # all_data = self.world_communicator.gather(pds.python_list, root=0)
        all_data = self.mpimanager.get_scheduler_communicator().gather(pds.python_list, root=0)

        # Initialize lists to accumulate results
        all_data_indices, all_data_items = [], []

        for node_data in all_data:
            for index, item in node_data:
                if isinstance(item, Exception):
                    raise item
                all_data_indices.append(index)
                all_data_items.append(item)

        # Sort the accumulated data according to the indices we tagged
        # them with when distributing
        rdd_sorted = [all_data_items[i] for i in np.argsort(all_data_indices)]
        return rdd_sorted

    def broadcast(self, value):
        """
        Sends a data to all leaders and workers
        First instruction is sent to leaders which then send it to their workers
        Then every process enters a broadcast to receive data from scheduler
        """
        # Tell the teams to enter broadcast()
        bds_id = self.__generate_new_bds_id()
        self.__command_teams(self.OP_BROADCAST, (bds_id,))

        _ = self.mpimanager.get_world_communicator().bcast(value, root=0)

        bds = BDSMPI(value, bds_id, self)
        return bds

    def delete_remote_pds(self, pds_id):
        """
        A public function for the PDS objects on the scheduler to call when they go out of
        scope or are deleted in order to ensure the same happens on the teams.

        Parameters
        ----------
        pds_id: int
            A pds_id identifying the remote PDS on the teams to delete.
        """

        if not self.finalized:
            self.__command_teams(self.OP_DELETEPDS, (pds_id,))

    def delete_remote_bds(self, bds_id):
        """
        Public function for the BDS objects on the scheduler to call when they go
        out of score or are deleted in order to ensure they are deleted
        ont he teams as well.

        Parameters
        ----------
        bds_id: int
            A bds_id identifying the remote BDS on the teams to delete.
        """

        if not self.finalized:
            # The scheduler deallocates it's BDS data. Explicit because
            # .. bds_store and BDSMPI object are disconnected.
            del backend.bds_store[bds_id]
            self.__command_teams(self.OP_DELETEBDS, (bds_id,))

    def __del__(self):
        """
        Overriding the delete function to explicitly call MPI.finalize().
        This is also required so we can tell the teams to get out of the
        while loop they are in and exit gracefully and they themselves call
        finalize when they die.
        """

        # Tell the teams they can exit gracefully.
        self.__command_teams(self.OP_FINISH, None)

        # Finalize the connection because the teams should have finished.
        MPI.Finalize()
        self.finalized = True


class BackendMPIWorker(Backend):
    """
    Workers are processes that are used to execute (maybe MPI) models
    There is one communicator per model to execute, compounded of one leader and workers
    Leaders receives instructions from the scheduler which then transmit them to workers
    Leaders are themselves workers 
    """

    OP_PARALLELIZE, OP_MAP, OP_COLLECT, OP_BROADCAST, OP_DELETEPDS, OP_DELETEBDS, OP_FINISH = [1, 2, 3, 4, 5, 6, 7]

    def __init__(self):
        """ No parameter, just call worker_run """
        self.logger = logging.getLogger(__name__)
        self.__worker_run()

    def run_function(self, function_packed, data_item):
        """
        Receives a serialized function unpack it and run it
        Passes the model communicator if ther is more than one process per model
        """
        func = cloudpickle.loads(function_packed)
        res = None
        try:
            if self.mpimanager.get_model_size() > 1:
                npc = NestedParallelizationControllerMPI(self.mpimanager.get_model_communicator())
                if self.mpimanager.get_model_communicator().Get_rank() == 0:
                    self.logger.debug("Executing map function on master rank 0.")
                    res = func(data_item, npc=npc)
                del npc
            else:
                res = func(data_item)
        except Exception as e:
            msg = "Exception occured while calling the map function {}: ".format(func.__name__)
            res = type(e)(msg + str(e))
        return res

    def __worker_run(self):
        """
        Workers enter an infinite loop and waits for instructions from their leader
        """
        while True:
            data = self.mpimanager.get_model_communicator().bcast(None, root=0)
            op = data[0]
            if op == self.OP_MAP:
                # Receive data from scheduler of the model
                function_packed = self.mpimanager.get_model_communicator().bcast(None, root=0)[0]
                data_item = self.mpimanager.get_model_communicator().bcast(None, root=0)[0]
                self.run_function(function_packed, data_item)
            elif op == self.OP_BROADCAST:
                self._bds_id = data[1]
                self.broadcast(None)
            elif op == self.OP_FINISH:
                quit()
            else:
                raise Exception("worker model received unknown command code")

    def collect(self):
        pass

    def map(self):
        pass

    def parallelize():
        pass

    def broadcast(self, value):
        """
        Receives data from scheduler
        """
        value = self.mpimanager.get_world_communicator().bcast(None, root=0)
        self.bds_store[self._bds_id] = value


class BackendMPILeader(BackendMPIWorker):
    """Defines the behavior of the leader processes

    This class defines how the leaders should behave during operation.
    leaders are those processes(not nodes like Spark) that have rank==0 in the model communicator
    """

    OP_PARALLELIZE, OP_MAP, OP_COLLECT, OP_BROADCAST, OP_DELETEPDS, OP_DELETEBDS, OP_FINISH = [1, 2, 3, 4, 5, 6, 7]

    def __init__(self):
        """ No parameter, just call leader_run """
        self.logger = logging.getLogger(__name__)
        self.__leader_run()

    def __leader_run(self):
        """
        This method is the infinite loop a leader enters directly from init.
        It makes the leader wait for a command to perform from the scheduler and
        then calls the appropriate function.

        This method also takes care of the synchronization of data between the
        scheduler and the leaders by matching PDSs based on the pds_ids sent by the scheduler
        with the command.

        Commands received from the scheduler are of the form of a tuple.
        The first component of the tuple is always the operation to be performed
        and the rest are conditional on the operation.

        (op,pds_id) where op == OP_PARALLELIZE for parallelize
        (op,pds_id, pds_id_result,func) where op == OP_MAP for map.
        (op,pds_id) where op == OP_COLLECT for a collect operation
        (op,pds_id) where op == OP_DELETEPDS for a delete of the remote PDS on leaders
        (op,) where op==OP_FINISH for the leader to break out of the loop and terminate
        """

        # Initialize PDS data store here because only teams need to do it.
        self.pds_store = {}

        while True:
            data = self.mpimanager.get_scheduler_communicator().bcast(None, root=0)

            op = data[0]
            if op == self.OP_PARALLELIZE:
                pds_id = data[1]
                self._rec_pds_id = pds_id
                pds_id, pds_id_new = self.__get_received_pds_id()
                self.pds_store[pds_id] = None


            elif op == self.OP_MAP:
                pds_id, pds_id_result, function_packed = data[1:]
                self._rec_pds_id, self._rec_pds_id_result = pds_id, pds_id_result

                # Enter the map so we can grab data and perform the func.
                # Func sent before and not during for performance reasons
                pds_res = self.map(function_packed)

                # Store the result in a newly generated PDS pds_id
                self.pds_store[pds_res.pds_id] = pds_res

            elif op == self.OP_BROADCAST:
                self._bds_id = data[1]
                # relay command and data into model communicator
                self.mpimanager.get_model_communicator().bcast(data, root=0)
                self.broadcast(None)

            elif op == self.OP_COLLECT:
                pds_id = data[1]

                # Access an existing PDS from data store
                pds = self.pds_store[pds_id]

                self.collect(pds)

            elif op == self.OP_DELETEPDS:
                pds_id = data[1]
                del self.pds_store[pds_id]

            elif op == self.OP_DELETEBDS:
                bds_id = data[1]
                del self.bds_store[bds_id]

            elif op == self.OP_FINISH:
                # tells other processes of the worker to finish
                self.mpimanager.get_model_communicator().bcast([self.OP_FINISH], root=0)
                quit()
            else:
                raise Exception("team received unknown command code")

    def __get_received_pds_id(self):
        """
        Function to retrieve the pds_id(s) we received from the scheduler to associate
        our team's created PDS with the scheduler's.
        """

        return self._rec_pds_id, self._rec_pds_id_result

    def __leader_run_function(self, function_packed, data_item):
        """
        This function sends data and serialized function to workers and executes it
        """
        self.mpimanager.get_model_communicator().bcast([self.OP_MAP], root=0)
        self.mpimanager.get_model_communicator().bcast([function_packed], root=0)
        self.mpimanager.get_model_communicator().bcast([data_item], root=0)
        return self.run_function(function_packed, data_item)

    def parallelize(self):
        pass

    def map(self, function_packed):
        """
        A distributed implementation of map that works on parallel data sets (PDS).
        On every element of pds the function func is called.
        We consider that process 0 of each MPI model should return the final result.

        Parameters
        ----------
        func: Python function_packed
            A serialized function that can be applied to every element of the pds

        Returns
        -------
        PDSMPI class
            a new parallel data set that contains the result of the map
        """

        map_start = time.time()

        # Get the PDS id we operate on and the new one to store the result in
        pds_id, pds_id_new = self.__get_received_pds_id()

        rdd = []
        while True:
            # Ask for a chunk of data since it's free
            data_chunks = self.mpimanager.get_scheduler_communicator().sendrecv(pds_id, 0, pds_id)

            # If it receives a sentinel, it's done and it can exit
            if data_chunks is None:
                break

            # Accumulate the indices and *processed* chunks
            for chunk in data_chunks:
                data_index, data_item = chunk
                res = self.__leader_run_function(function_packed, data_item)
                rdd += [(data_index, res)]

        pds_res = PDSMPI(rdd, pds_id_new, self)

        return pds_res

    def collect(self, pds):
        """
        Gather the pds from all the leaders,
        send it to the scheduler and return it as a standard Python list.

        Parameters
        ----------
        pds: PDS class
            a parallel data set

        Returns
        -------
        Python list
            all elements of pds as a list
        """

        # Send the data we have back to the scheduler
        _ = self.mpimanager.get_scheduler_communicator().gather(pds.python_list, root=0)


class BackendMPITeam(BackendMPILeader if abcpy.backends.mpimanager.get_mpi_manager().is_leader() else BackendMPIWorker):
    """
    A team is compounded by workers and a leader. One process per team is a leader, others are workers
    """

    OP_PARALLELIZE, OP_MAP, OP_COLLECT, OP_BROADCAST, OP_DELETEPDS, OP_DELETEBDS, OP_FINISH = [1, 2, 3, 4, 5, 6, 7]

    def __init__(self):
        # Define the vars that will hold the pds ids received from scheduler to operate on
        self._rec_pds_id = None
        self._rec_pds_id_result = None

        # Initialize a BDS store for both scheduler & team.
        self.bds_store = {}

        # print("In BackendMPITeam, rank : ", self.rank, ", model_rank_global : ", globals()['model_rank_global'])

        self.logger = logging.getLogger(__name__)
        super().__init__()


class BackendMPI(BackendMPIScheduler if abcpy.backends.mpimanager.get_mpi_manager().is_scheduler() else BackendMPITeam):
    """A backend parallelized by using MPI

    The backend conditionally inherits either the BackendMPIScheduler class
    or the BackendMPIteam class depending on it's rank. This lets
    BackendMPI have a uniform interface for the user but allows for a
    logical split between functions performed by the scheduler
    and the teams.
    """

    def __init__(self, scheduler_node_ranks=[0], process_per_model=1):
        """
        Parameters
        ----------
        scheduler_node_ranks: Python list
            list of scheduler nodes
        
        process_per_model: Integer
            number of MPI processes to allocate to each model
        """
        # get mpimanager instance from the mpimanager module (which has to be setup before calling the constructor)
        self.logger = logging.getLogger(__name__)
        self.mpimanager = abcpy.backends.mpimanager.get_mpi_manager()

        if self.mpimanager.get_world_size() < 2:
            raise ValueError('A minimum of 2 ranks are required for the MPI backend')

        # Set the global backend
        globals()['backend'] = self

        # Call the appropriate constructors and pass the required data
        super().__init__()

    def size(self):
        """ Returns world size """
        return self.mpimanager.get_world_size()

    def scheduler_node_ranks(self):
        """ Returns scheduler node ranks """
        return self.mpimanager.get_scheduler_node_ranks()

    @staticmethod
    def disable_nested(mpi_comm):
        if mpi_comm.Get_rank() != 0:
            mpi_comm.Barrier()

    @staticmethod
    def enable_nested(mpi_comm):
        if mpi_comm.Get_rank() == 0:
            mpi_comm.Barrier()


class PDSMPI(PDS):
    """
    This is an MPI wrapper for a Python parallel data set.
    """

    def __init__(self, python_list, pds_id, backend_obj):
        self.python_list = python_list
        self.pds_id = pds_id
        self.backend_obj = backend_obj

    def __del__(self):
        """
        Destructor to be called when a PDS falls out of scope and/or is being deleted.
        Uses the backend to send a message to destroy the teams' copy of the pds.
        """
        try:
            self.backend_obj.delete_remote_pds(self.pds_id)
        except AttributeError:
            # Catch "delete_remote_pds not defined" for teams and ignore.
            pass


class BDSMPI(BDS):
    """
    This is a wrapper for MPI's BDS class.
    """

    def __init__(self, object, bds_id, backend_obj):
        # The BDS data is no longer saved in the BDS object.
        # It will access & store the data only from the current backend
        self.bds_id = bds_id
        backend.bds_store[self.bds_id] = object

    def value(self):
        """
        This method returns the actual object that the broadcast data set represents.
        """
        return backend.bds_store[self.bds_id]

    def __del__(self):
        """
        Destructor to be called when a BDS falls out of scope and/or is being deleted.
        Uses the backend to send a message to destroy the teams' copy of the bds.
        """

        try:
            backend.delete_remote_bds(self.bds_id)
        except AttributeError:
            # Catch "delete_remote_pds not defined" for teams and ignore.
            pass


class BackendMPITestHelper:
    """
    Helper function for some of the test cases to be able to access and verify class members.
    """

    def check_pds(self, k):
        """Checks if a PDS exists in the pds data store. Used to verify deletion and creation
        """
        return k in backend.pds_store.keys()

    def check_bds(self, k):
        """Checks if a BDS exists in the BDS data store. Used to verify deletion and creation
        """
        return k in backend.bds_store.keys()



import unittest

from abcpy.continuousmodels import *
from abcpy.probabilisticmodels import *

"""These test cases implement tests for the probabilistic model class."""


class AbstractAPIImplementationTests:
    def setUp(self):
        self.models = []
        for model_type, model_input in zip(self.model_types, self.model_inputs):
            model = model_type(model_input)
            self.models.append(model)

    def test__getitem__(self):
        for model in self.models:
            item = model[0]
            self.assertTrue(isinstance(item, InputConnector),
                            'Return value not of type InputConnector for model {}.'.format(type(model)))

    def test_get_input_values(self):
        for model in self.models:
            values = model.get_input_values()
            self.assertTrue(isinstance(values, list), 'Return value not of type list in model {}.'.format(type(model)))
            self.assertEqual(len(values), model.get_input_dimension(),
                             'Number of parameters not equal to input dimension of model {}.'.format(model))

    def test_get_input_models(self):
        for model in self.models:
            in_models = model.get_input_models()
            self.assertTrue(isinstance(in_models, list),
                            'Return value not of type list in model {}.'.format(type(model)))
            self.assertEqual(len(in_models), model.get_input_dimension(),
                             'Number of parameters not equal to input dimension of model {}.'.format(model))

    def test_get_stored_output_values(self):
        for model in self.models:
            rng = np.random.RandomState(1)
            model._forward_simulate_and_store_output(rng)
            out_values = model.get_stored_output_values()
            self.assertTrue(isinstance(out_values, np.ndarray),
                            'Return value not of type numpy.array in model {}.'.format(type(model)))
            self.assertEqual(len(out_values), model.get_output_dimension(),
                             'Number of parameters not equal to output dimension of model {}.'.format(model))

    def test_get_input_connector(self):
        for model in self.models:
            in_con = model.get_input_connector()
            self.assertTrue(isinstance(in_con, InputConnector) or in_con == None,
                            'Return value not of type InputConnector nor None in model {}.'.format(type(model)))

    def test_get_input_dimension(self):
        for model in self.models:
            dim = model.get_input_dimension()
            self.assertTrue(isinstance(dim, Number), 'Return value not of type Number in model {}.'.format(type(model)))
            self.assertGreaterEqual(dim, 0, 'Input dimension must be larger than 0 for model {}.'.format(type(model)))

    def test_set_output_values(self):
        for model in self.models:
            number = 1
            with self.assertRaises(TypeError) as context:
                model.set_output_values(number)
            self.assertTrue(context.exception, 'Model {} should not accept a number as input.'.format(type(model)))

            nparray = np.ones(model.get_output_dimension() + 1)
            with self.assertRaises(IndexError) as context:
                model.set_output_values(nparray)
            self.assertTrue(context.exception,
                            'Model {} should only accept input equal to output dimension.'.format(type(model)))

    def test_pdf(self):
        for model in self.models:
            x = 0
            input = model.get_input_values()
            pdf_at_x = model.pdf(input, x)
            self.assertTrue(isinstance(pdf_at_x, Number),
                            'Return value not of type Number in model {}.'.format(type(model)))

    def test_check_input(self):
        for model in self.models:
            test_result = model._check_input(model.get_input_values())
            self.assertTrue(test_result,
                            'The checking method should return True if input is reasonable in model {}.'.format(
                                type(model)))

            with self.assertRaises(Exception) as context:
                model._check_input(0)
            self.assertTrue(context.exception,
                            'Function should raise an exception in model {} if input not of type InputConnector.'.format(
                                type(model)))

    def test_forward_simulate(self):
        for model in self.models:
            rng = np.random.RandomState(1)
            result_list = model.forward_simulate(model.get_input_values(), 3, rng)
            self.assertTrue(isinstance(result_list, list),
                            'Return value not of type list in model {}.'.format(type(model)))
            self.assertEqual(len(result_list), 3,
                             'Model {} did not return the requseted number of formard simulations.'.format(type(model)))

            result = result_list[0]
            self.assertTrue(isinstance(result, np.ndarray),
                            'A single forward simulation is not of type numpy.array in model {}.'.format(type(model)))

    def test_get_output_dimension(self):
        for model in self.models:
            expected_dim = model.get_output_dimension()
            self.assertTrue(isinstance(expected_dim, Number),
                            'Return value not of type Number in model {}.'.format(type(model)))
            self.assertGreater(expected_dim, 0,
                               'Output dimension must be larger than 0 for model {}.'.format(type(model)))

            rng = np.random.RandomState(1)
            result_list = model.forward_simulate(model.get_input_values(), 1, rng)
            result = result_list[0]
            result_dim = result.shape[0]
            self.assertEqual(result_dim, expected_dim,
                             'Output dimension of forward simulation is not equal to get_output_dimension() for model {}.'.format(
                                 type(model)))


class HyperParameterAPITests(AbstractAPIImplementationTests, unittest.TestCase):
    model_types = [Hyperparameter]
    model_inputs = [1]


class GetItemTest(unittest.TestCase):
    """Tests whether the get_item operator (i.e. access operator) works as intended."""

    def ErrorTest(self):
        """Tests whether an error is thrown if the operator tries to access an element outside the range of the return values of the probabilistic model."""
        N = Normal([1, 0.1], )
        with self.assertRaises(IndexError):
            N[2]

    def ReturnValueTest(self):
        """Tests whether the return value of the operator is correct"""
        N = Normal([1, 0.1], )
        result = N[0]
        self.assertTrue(isinstance(result, tuple))
        self.assertTrue(result[0] == N)
        self.assertTrue(result[1] == 0)


class MappingTest(unittest.TestCase):
    """Tests whether the mapping created during initialization is done correctly."""

    def setUp(self):
        self.U = Uniform([[0, 2], [1, 3]])
        self.M = MultivariateNormal([[self.U[1], self.U[0]], [[1, 0], [0, 1]]])

    def test(self):
        self.assertTrue(self.M.get_input_connector().get_model(0) == self.U)
        self.assertTrue(self.M.get_input_connector().get_model(1) == self.U)


class GetParameterValuesTest(unittest.TestCase):
    """Tests whether get_input_values returns the correct values."""

    def test(self):
        U = Uniform([[0, 2], [1, 3]])
        self.assertTrue(U.get_input_values() == [0, 2, 1, 3])


class SampleParametersTest(unittest.TestCase):
    """Tests whether _forward_simulate_and_store_output returns False if the value of an input parameter is not an allowed value for the
    distribution."""

    def test(self):
        N1 = Normal([0.1, 0.01])
        N2 = Normal([1, N1])
        N1._fixed_values = [-0.1]
        self.assertFalse(N2._check_input(N2.get_input_values()))


class GetOutputValuesTest(unittest.TestCase):
    """Tests whether get_stored_output_values gives back values that can come from the distribution."""

    def test(self):
        U = Uniform([[0], [1]])
        U._forward_simulate_and_store_output()
        self.assertTrue((U.get_stored_output_values()[0] >= 0 and U.get_stored_output_values()[0] <= 1))


class ModelResultingFromOperationTests(unittest.TestCase):

    def test_check_parameters_at_initialization(self):
        N1 = Normal([1, 0.1])
        M1 = MultivariateNormal([[1, 1], [[1, 0], [0, 1]]])
        with self.assertRaises(ValueError):
            model = N1 + M1

        def test_initialization(self):
            M1 = MultivariateNormal([[1, 1], [[1, 0], [0, 1]]])
            M2 = MultivariateNormal([[1, 1], [[1, 0], [0, 1]]])
            M3 = M1 + M2
            self.assertTrue(M3.get_output_dimension() == 2)


class SummationModelTests(unittest.TestCase):
    """Tests whether all methods associated with the SummationModel are working as intended."""

    def test_forward_simulate(self):
        N1 = Normal([1, 0.1])
        N2 = 10 + N1
        rng = np.random.RandomState(1)
        N1._forward_simulate_and_store_output(rng=rng)

        sample = N2.forward_simulate(N2.get_input_values(), 1, rng)

        self.assertTrue(isinstance(sample[0], np.ndarray))


class SubtractionModelTests(unittest.TestCase):
    """Tests whether all methods associated with the SubtractionModel are working as intended."""

    def test_forward_simulate(self):
        N1 = Normal([1, 0.1])
        N2 = 10 - N1
        rng = np.random.RandomState(1)
        N1._forward_simulate_and_store_output(rng=rng)

        sample = N2.forward_simulate(N2.get_input_values(), 1, rng)

        self.assertTrue(isinstance(sample[0], np.ndarray))


class MultiplicationModelTests(unittest.TestCase):
    """Tests whether all methods associated with the MultiplicationModel are working as intended."""

    def test_forward_simulate(self):
        N1 = Normal([1, 0.1])
        N2 = N1 * 2
        rng = np.random.RandomState(1)
        N1._forward_simulate_and_store_output(rng=rng)

        sample = N2.forward_simulate(N2.get_input_values(), 1, rng)
        self.assertTrue(isinstance(sample[0], np.ndarray))

    def test_multiplication_from_right(self):
        N1 = Normal([1, 0.1])
        N2 = 2 * N1

        self.assertTrue(N2.get_input_dimension() == 2)
        self.assertTrue(isinstance(N2.get_input_connector().get_model(0), Hyperparameter))


class DivisionModelTests(unittest.TestCase):
    """Tests whether all methods associated with the DivisionModel are working as intended."""

    def test_sample_from_distribution(self):
        N1 = Normal([1, 0.1])
        N2 = Normal([2, 0.1])
        N3 = N1 / N2
        rng = np.random.RandomState(1)

        N1._forward_simulate_and_store_output(rng=rng)
        N2._forward_simulate_and_store_output(rng=rng)
        sample = N3.forward_simulate(N3.get_input_values(), 1)

        self.assertTrue(isinstance(sample[0], np.ndarray))

    def test_division_from_right(self):
        N1 = Normal([1, 0.1])
        N2 = 2 / N1

        self.assertEqual(N2.get_input_dimension(), 2)
        self.assertTrue(isinstance(N2.get_input_connector().get_model(0), Hyperparameter))


class ExponentialModelTests(unittest.TestCase):
    """Tests whether all methods associated with ExponentialModel are working as intended."""

    def test_check_parameters_at_initialization(self):
        """Tests whether it is possible to have a multidimensional exponent."""
        M1 = MultivariateNormal([[1, 1], [[1, 0], [0, 1]]])
        N1 = Normal([1, 0.1])
        with self.assertRaises(ValueError):
            N1 ** M1

    def test_initialization(self):
        """Tests that no errors during initialization are raised."""
        N1 = Normal([1, 0.1])
        N2 = Normal([1, 0.1])

        N3 = N1 ** N2

        N4 = N1 ** 2

        N5 = 2 ** N1

    def test_forward_simulate(self):
        """Tests whether forward_simulate gives the desired output."""
        N1 = Normal([1, 0.1])
        N2 = N1 ** 2
        rng = np.random.RandomState(1)
        N1._forward_simulate_and_store_output(rng=rng)
        sample = N2.forward_simulate(N2.get_input_values(), 1, rng=rng)
        self.assertTrue(isinstance(sample[0], np.ndarray))


if __name__ == '__main__':
    unittest.main()



from functools import wraps

import numpy as np
import ot


def cached(func):
    cache = {}

    @wraps(func)
    def wrapped(x):
        if x not in cache:
            cache[x] = func(x)
        return cache[x]

    return wrapped


def wass_dist(samples_1, samples_2, weights_1=None, weights_2=None, num_iter_max=100000, **kwargs):
    """
    Computes the Wasserstein 2 distance between two empirical distributions with weights. This uses the POT library to 
    estimate Wasserstein distance. The Wasserstein distance computation can take long if the number of samples in the 
    two datasets is large (cost of the computation scales in fact quadratically with the number of samples).

    Parameters
    ----------
    samples_1 : np.ndarray
         Samples defining the first empirical distribution, with shape (nxd), n being the number of samples in the
         first empirical distribution and d the dimension of the random variable.
    samples_2 : np.ndarray
         Samples defining the second empirical distribution, with shape (mxd), m being the number of samples in the
         second empirical distribution and d the dimension of the random variable.
    weights_1 : np.ndarray, optional
         Weights defining the first empirical distribution, with shape (n), n being the number of samples in the
         first empirical distribution. Weights are normalized internally to the function. If not provided, they are
         assumed to be identical for all samples.
    weights_2 : np.ndarray, optional
         Weights defining the second empirical distribution, with shape (m), m being the number of samples in the
         second empirical distribution. Weights are normalized internally to the function. If not provided, they are
         assumed to be identical for all samples.
    num_iter_max : integer, optional
        The maximum number of iterations in the linear programming algorithm to estimate the Wasserstein distance. 
        Default to 100000. 
    kwargs 
        Additional arguments passed to ot.emd2

    Returns
    -------
    float
        The estimated 2-Wasserstein distance.
    """
    n = samples_1.shape[0]
    m = samples_2.shape[0]

    if weights_1 is None:
        a = np.ones((n,)) / n
    else:
        if len(weights_1) != n:
            raise RuntimeError("Number of weights and number of samples need to be the same.")
        a = weights_1 / np.sum(weights_1)
    if weights_2 is None:
        b = np.ones((m,)) / m
    else:
        if len(weights_2) != m:
            raise RuntimeError("Number of weights and number of samples need to be the same.")
        b = weights_2 / np.sum(weights_2)

    # loss matrix
    M = ot.dist(x1=samples_1, x2=samples_2)  # this returns squared distance!
    cost = ot.emd2(a, b, M, numItermax=num_iter_max, **kwargs)

    return np.sqrt(cost)



import unittest

import numpy as np

from abcpy.approx_lhd import SynLikelihood
from abcpy.backends import BackendDummy
from abcpy.continuousmodels import Normal
from abcpy.continuousmodels import Uniform
from abcpy.distances import Euclidean, MMD
from abcpy.inferences import RejectionABC, PMC, PMCABC, SABC, ABCsubsim, SMCABC, APMCABC, RSMCABC, \
    MCMCMetropoliHastings
from abcpy.statistics import Identity


class RejectionABCTest(unittest.TestCase):
    def setUp(self):
        # setup backend
        dummy = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define sufficient statistics for the model
        stat_calc = Identity(degree=2, cross=False)

        # define a distance function
        dist_calc = Euclidean(stat_calc)

        # create fake observed data
        self.y_obs = [np.array(9.8)]

        # for correct seeding define 2 samplers
        self.sampler = RejectionABC([self.model], [dist_calc], dummy, seed=1)
        self.sampler2 = RejectionABC([self.model], [dist_calc], dummy, seed=1)

    def test_sample_n_samples(self):
        # use the rejection sampling scheme
        journal = self.sampler.sample([self.y_obs], 10, 1, 10, path_to_save_journal="tmp.jnl")
        mu_sample = np.array(journal.get_parameters()['mu'])
        sigma_sample = np.array(journal.get_parameters()['sigma'])

        # test shape of samples
        mu_shape, sigma_shape = (len(mu_sample), mu_sample[0].shape[1]), \
                                (len(sigma_sample), sigma_sample[0].shape[1])
        self.assertEqual(mu_shape, (10, 1))
        self.assertEqual(sigma_shape, (10, 1))

        # Compute posterior mean
        # self.assertAlmostEqual(np.average(np.asarray(samples[:,0])),1.22301,10e-2)
        self.assertAlmostEqual(np.average(mu_sample), 1.223012836345375)
        self.assertAlmostEqual(np.average(sigma_sample), 6.992218962395242)

        self.assertFalse(journal.number_of_simulations[0] == 0)
        self.assertEqual(journal.configuration["epsilon"], 10)

    def test_sample_simulation_budget(self):
        # use the rejection sampling scheme with epsilon first
        journal = self.sampler2.sample([self.y_obs], n_samples=None, simulation_budget=100, epsilon=20)
        mu_sample = np.array(journal.get_parameters()['mu'])
        sigma_sample = np.array(journal.get_parameters()['sigma'])

        mu_shape, sigma_shape = (len(mu_sample), mu_sample[0].shape[1]), \
                                (len(sigma_sample), sigma_sample[0].shape[1])
        self.assertEqual(mu_shape, (3, 1))
        self.assertEqual(sigma_shape, (3, 1))

        # Compute posterior mean
        # self.assertAlmostEqual(np.average(np.asarray(samples[:,0])),1.22301,10e-2)
        self.assertAlmostEqual(np.average(mu_sample), 0.8175361535037666)
        self.assertAlmostEqual(np.average(sigma_sample), 8.155647092489977)

        self.assertEqual(journal.number_of_simulations[0], 100)
        self.assertEqual(journal.configuration["epsilon"], 20)

        # use the rejection sampling scheme with the quantile
        journal = self.sampler2.sample([self.y_obs], n_samples=None, simulation_budget=100, quantile=0.1)
        mu_sample = np.array(journal.get_parameters()['mu'])
        sigma_sample = np.array(journal.get_parameters()['sigma'])

        mu_shape, sigma_shape = (len(mu_sample), mu_sample[0].shape[1]), \
                                (len(sigma_sample), sigma_sample[0].shape[1])
        self.assertEqual(mu_shape, (10, 1))
        self.assertEqual(sigma_shape, (10, 1))

        # Compute posterior mean
        # self.assertAlmostEqual(np.average(np.asarray(samples[:,0])),1.22301,10e-2)
        self.assertAlmostEqual(np.average(mu_sample), 0.10394992719538543)
        self.assertAlmostEqual(np.average(sigma_sample), 6.746940834914168)

        self.assertEqual(journal.number_of_simulations[0], 200)

    def test_errors(self):
        with self.assertRaises(RuntimeError):
            self.sampler.sample([self.y_obs], n_samples=10, simulation_budget=10)
        with self.assertRaises(RuntimeError):
            self.sampler.sample([self.y_obs], n_samples=10, quantile=0.1)
        with self.assertRaises(RuntimeError):
            self.sampler.sample([self.y_obs], n_samples=10, epsilon=None)
        with self.assertRaises(RuntimeError):
            self.sampler.sample([self.y_obs], n_samples=None, simulation_budget=100, quantile=0.1, epsilon=1)
        with self.assertRaises(RuntimeError):
            self.sampler.sample([self.y_obs], n_samples=None, simulation_budget=100, quantile=None, epsilon=None)


class PMCTests(unittest.TestCase):

    def test_sample(self):
        # setup backend
        backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define sufficient statistics for the model
        stat_calc = Identity(degree=2, cross=False)

        # create fake observed data
        # y_obs = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        y_obs = [np.array(9.8)]

        # Define the likelihood function
        likfun = SynLikelihood(stat_calc)

        T, n_sample, n_samples_per_param = 1, 10, 100
        sampler = PMC([self.model], [likfun], backend, seed=1)
        journal = sampler.sample([y_obs], T, n_sample, n_samples_per_param, covFactors=np.array([.1, .1]),
                                 iniPoints=None, path_to_save_journal="tmp.jnl")
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -3.3711206204663773, delta=1e-3)
        self.assertAlmostEqual(sigma_post_mean, 6.519325027532673, delta=1e-3)

        self.assertFalse(journal.number_of_simulations == 0)

        # use the PMC scheme for T = 2
        T, n_sample, n_samples_per_param = 2, 10, 100
        sampler = PMC([self.model], [likfun], backend, seed=1)
        journal = sampler.sample([y_obs], T, n_sample, n_samples_per_param, covFactors=np.array([.1, .1]),
                                 iniPoints=None)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -3.2517600952705257, delta=1e-3)
        self.assertAlmostEqual(sigma_post_mean, 6.9214661382633365, delta=1e-3)
        self.assertFalse(journal.number_of_simulations == 0)


class MCMCMetropoliHastingsTests(unittest.TestCase):
    # test if MCMCMetropoliHastings works

    def setUp(self):
        # setup backend
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])
        self.model2 = Normal([mu, sigma])

        # define sufficient statistics for the model
        stat_calc = Identity(degree=2, cross=False)

        # create fake observed data
        # y_obs = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.y_obs = [np.array(9.8)]
        self.y_obs2 = [np.array(3.4)]

        # Define the likelihood function
        self.likfun = SynLikelihood(stat_calc)
        self.likfun2 = SynLikelihood(stat_calc)

        self.bounds = {"mu": [-5, 5], "sigma": (0, 10)}

    def test_sample(self):
        n_sample, n_samples_per_param = 50, 20

        sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
        journal = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                 iniPoint=None, burnin=10, adapt_proposal_cov_interval=5, use_tqdm=False,
                                 path_to_save_journal="tmp.jnl")
        # without speedup_dummy
        sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)  # to reset seed
        journal_repr = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None, use_tqdm=False,
                                      iniPoint=None, speedup_dummy=False, burnin=10, adapt_proposal_cov_interval=5)
        # Compute posterior mean
        mu_post_mean_1, sigma_post_mean_1 = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']
        mu_post_mean_2, sigma_post_mean_2 = journal_repr.posterior_mean()['mu'], journal_repr.posterior_mean()['sigma']

        self.assertAlmostEqual(mu_post_mean_1, 3.735372456561986)
        self.assertAlmostEqual(mu_post_mean_2, -0.6946660151693353)
        self.assertAlmostEqual(sigma_post_mean_1, 5.751158868437219)
        self.assertAlmostEqual(sigma_post_mean_2, 8.103358539327967)

    def test_sample_with_transformer(self):
        n_sample, n_samples_per_param = 50, 20

        sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
        journal = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                 iniPoint=None, burnin=10, adapt_proposal_cov_interval=5, use_tqdm=False,
                                 bounds=self.bounds)
        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        self.assertAlmostEqual(mu_post_mean, 1.3797371606192235)
        self.assertAlmostEqual(sigma_post_mean, 8.097776586316062)

        # test raises correct errors:
        with self.assertRaises(TypeError):
            journal = sampler.sample([self.y_obs], n_sample, bounds=[0, 1])
        with self.assertRaises(KeyError):
            journal = sampler.sample([self.y_obs], n_sample, bounds={"hello": (0, 1)})
        with self.assertRaises(RuntimeError):
            journal = sampler.sample([self.y_obs], n_sample, bounds={"mu": (0)})
        with self.assertRaises(RuntimeError):
            journal = sampler.sample([self.y_obs], n_sample, bounds={"mu": (0, 1, 2)})

    def test_sample_two_models(self):
        n_sample, n_samples_per_param = 50, 20

        sampler = MCMCMetropoliHastings([self.model, self.model2], [self.likfun, self.likfun2], self.backend,
                                        seed=1)
        journal = sampler.sample([self.y_obs, self.y_obs2], n_sample, n_samples_per_param, cov_matrices=None,
                                 iniPoint=None, burnin=10, adapt_proposal_cov_interval=5, use_tqdm=False)
        # without speedup_dummy
        sampler = MCMCMetropoliHastings([self.model, self.model2], [self.likfun, self.likfun2], self.backend,
                                        seed=1)  # to reset seed
        journal_repr = sampler.sample([self.y_obs, self.y_obs2], n_sample, n_samples_per_param, cov_matrices=None,
                                      iniPoint=None, speedup_dummy=False, burnin=10, adapt_proposal_cov_interval=5,
                                      use_tqdm=False)
        # Compute posterior mean
        mu_post_mean_1, sigma_post_mean_1 = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']
        mu_post_mean_2, sigma_post_mean_2 = journal_repr.posterior_mean()['mu'], journal_repr.posterior_mean()['sigma']

        self.assertAlmostEqual(mu_post_mean_1, 0.1920594166217264)
        self.assertAlmostEqual(mu_post_mean_2, -1.0095854412936525)
        self.assertAlmostEqual(sigma_post_mean_1, 9.143353645946233)
        self.assertAlmostEqual(sigma_post_mean_2, 7.539268611159257)

    def test_restart_from_journal(self):
        for speedup_dummy in [True, False]:
            # do at once:
            n_sample, n_samples_per_param = 40, 20
            sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
            journal_at_once = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                             iniPoint=None, speedup_dummy=speedup_dummy, burnin=20,
                                             adapt_proposal_cov_interval=10, use_tqdm=False)
            # do separate:
            n_sample, n_samples_per_param = 20, 20
            sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
            journal = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None, iniPoint=None,
                                     speedup_dummy=speedup_dummy, burnin=20, adapt_proposal_cov_interval=10,
                                     use_tqdm=False, path_to_save_journal="tmp.jnl")

            journal_separate = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                              iniPoint=None, journal_file="tmp.jnl", burnin=0,
                                              speedup_dummy=speedup_dummy, use_tqdm=False)  # restart from this journal

            self.assertEqual(journal_separate.configuration['n_samples'], journal_at_once.configuration['n_samples'])
            self.assertEqual(journal_separate.number_of_simulations[-1], journal_at_once.number_of_simulations[-1])
            self.assertEqual(journal_separate.configuration["acceptance_rates"][-1],
                             journal_at_once.configuration["acceptance_rates"][-1])
            self.assertEqual(len(journal_separate.get_parameters()), len(journal_at_once.get_parameters()))
            self.assertEqual(len(journal_separate.get_parameters()['mu']), len(journal_at_once.get_parameters()['mu']))
            self.assertEqual(len(journal_separate.get_accepted_parameters()),
                             len(journal_at_once.get_accepted_parameters()))
            self.assertEqual(journal_separate.get_weights().shape, journal_at_once.get_weights().shape)
            self.assertEqual(journal_separate.posterior_mean()['mu'], journal_at_once.posterior_mean()['mu'])

    def test_restart_from_journal_with_transformer(self):
        for speedup_dummy in [True, False]:
            # do at once:
            n_sample, n_samples_per_param = 40, 20
            sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
            journal_at_once = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                             iniPoint=None, speedup_dummy=speedup_dummy, burnin=20,
                                             adapt_proposal_cov_interval=10, use_tqdm=False, bounds=self.bounds)
            # do separate:
            n_sample, n_samples_per_param = 20, 20
            sampler = MCMCMetropoliHastings([self.model], [self.likfun], self.backend, seed=1)
            journal = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None, iniPoint=None,
                                     speedup_dummy=speedup_dummy, burnin=20, adapt_proposal_cov_interval=10,
                                     use_tqdm=False, bounds=self.bounds, path_to_save_journal="tmp.jnl")

            journal_separate = sampler.sample([self.y_obs], n_sample, n_samples_per_param, cov_matrices=None,
                                              iniPoint=None, journal_file="tmp.jnl", burnin=0,
                                              speedup_dummy=speedup_dummy, use_tqdm=False,
                                              bounds=self.bounds)  # restart from this journal

            self.assertEqual(journal_separate.configuration['n_samples'], journal_at_once.configuration['n_samples'])
            self.assertEqual(journal_separate.number_of_simulations[-1], journal_at_once.number_of_simulations[-1])
            self.assertEqual(journal_separate.configuration["acceptance_rates"][-1],
                             journal_at_once.configuration["acceptance_rates"][-1])
            self.assertEqual(len(journal_separate.get_parameters()), len(journal_at_once.get_parameters()))
            self.assertEqual(len(journal_separate.get_parameters()['mu']), len(journal_at_once.get_parameters()['mu']))
            self.assertEqual(len(journal_separate.get_accepted_parameters()),
                             len(journal_at_once.get_accepted_parameters()))
            self.assertEqual(journal_separate.get_weights().shape, journal_at_once.get_weights().shape)
            self.assertAlmostEqual(journal_separate.posterior_mean()['mu'], journal_at_once.posterior_mean()['mu'])


class PMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        # self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

    def test_calculate_weight(self):
        n_samples = 2
        rc = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        theta = np.array([1.0, 1.0])

        weight = rc._calculate_weight(theta)
        self.assertEqual(weight, 0.5)

        accepted_parameters = [[1.0, 1.0 + np.sqrt(2)], [0, 0]]
        accepted_weights = np.array([[.5], [.5]])
        accepted_cov_mat = [np.array([[1.0, 0], [0, 1]])]
        rc.accepted_parameters_manager.update_broadcast(rc.backend, accepted_parameters, accepted_weights,
                                                        accepted_cov_mat)
        kernel_parameters = []
        for kernel in rc.kernel.kernels:
            kernel_parameters.append(
                rc.accepted_parameters_manager.get_accepted_parameters_bds_values(kernel.models))

        rc.accepted_parameters_manager.update_kernel_values(rc.backend, kernel_parameters=kernel_parameters)
        weight = rc._calculate_weight(theta)
        expected_weight = 0.170794684453
        self.assertAlmostEqual(weight, expected_weight)

    def test_sample(self):
        # use the PMCABC scheme for T = 1
        T, n_sample, n_simulate, eps_arr, eps_percentile = 1, 10, 1, [10], 10
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], T, eps_arr, n_sample, n_simulate, eps_percentile)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - 0.03713, 10e-2)
        self.assertLess(sigma_post_mean - 7.727, 10e-2)

        # self.assertEqual((mu_post_mean, sigma_post_mean), (,))

        # use the PMCABC scheme for T = 2
        T, n_sample, n_simulate, eps_arr, eps_percentile = 2, 10, 1, [10, 5], 10
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal = sampler.sample([self.observation], T, eps_arr, n_sample, n_simulate, eps_percentile)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - 0.9356, 10e-2)
        self.assertLess(sigma_post_mean - 7.819, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

        # use the PMCABC scheme for T = 2 providing only first value for eps_arr
        T, n_sample, n_simulate, eps_arr, eps_percentile = 2, 10, 1, [10], 10
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal = sampler.sample([self.observation], T, eps_arr, n_sample, n_simulate, eps_percentile)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - 0.9356, 10e-2)
        self.assertLess(sigma_post_mean - 7.819, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

    def test_restart_from_journal(self):
        # test with value of eps_arr_2 > percentile of distances
        n_sample, n_simulate, eps_arr, eps_percentile = 10, 1, [10, 5], 10
        # 2 steps with intermediate journal:
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal_intermediate = sampler.sample([self.observation], 1, [eps_arr[0]], n_sample, n_simulate, eps_percentile,
                                              path_to_save_journal="tmp.jnl")

        journal_final_1 = sampler.sample([self.observation], 1, [eps_arr[1]], n_sample, n_simulate, eps_percentile,
                                         journal_file="tmp.jnl")
        # 2 steps directly
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal_final_2 = sampler.sample([self.observation], 2, eps_arr, n_sample, n_simulate, eps_percentile)

        self.assertEqual(journal_final_1.configuration["epsilon_arr"], journal_final_2.configuration["epsilon_arr"])
        self.assertEqual(journal_final_1.posterior_mean()['mu'], journal_final_2.posterior_mean()['mu'])

        # test with value of eps_arr_2 < percentile of distances
        n_sample, n_simulate, eps_arr, eps_percentile = 10, 1, [10, 1], 10
        # 2 steps with intermediate journal:
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal_intermediate = sampler.sample([self.observation], 1, [eps_arr[0]], n_sample, n_simulate, eps_percentile)
        journal_intermediate.save("tmp.jnl")
        journal_final_1 = sampler.sample([self.observation], 1, [eps_arr[1]], n_sample, n_simulate, eps_percentile,
                                         journal_file="tmp.jnl")
        # 2 steps directly
        sampler = PMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal_final_2 = sampler.sample([self.observation], 2, eps_arr, n_sample, n_simulate, eps_percentile)

        self.assertEqual(journal_final_1.configuration["epsilon_arr"], journal_final_2.configuration["epsilon_arr"])
        self.assertEqual(journal_final_1.posterior_mean()['mu'], journal_final_2.posterior_mean()['mu'])


class SABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)
        self.dist_calc_mmd = MMD(stat_calc, biased_estimator=False)

        # create fake observed data
        # self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

    def test_sample(self):
        # use the SABC scheme for T = 1
        steps, epsilon, n_samples, n_samples_per_param = 1, 10, 10, 1
        sampler = SABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, epsilon, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[0]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))

        # use the SABC scheme for T = 2
        steps, epsilon, n_samples, n_samples_per_param = 2, 10, 10, 1
        sampler = SABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, epsilon, n_samples, n_samples_per_param, full_output=1,
                                 path_to_save_journal="tmp")
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - 0.55859197, 10e-2)
        self.assertLess(sigma_post_mean - 7.03987723, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)

        # check whether it raises the correct error with MMD:
        with self.assertRaises(RuntimeError):
            sampler = SABC([self.model], [self.dist_calc_mmd], self.backend, seed=1)


class ABCsubsimTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        # self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

    def test_sample(self):
        # use the ABCsubsim scheme for T = 1
        steps, n_samples, n_samples_per_param = 1, 10, 1
        sampler = ABCsubsim([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))

        # use the ABCsubsim scheme for T = 2
        steps, n_samples, n_samples_per_param = 2, 10, 1
        sampler = ABCsubsim([self.model], [self.dist_calc], self.backend, seed=1)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        journal = sampler.sample([self.observation], steps, n_samples, n_samples_per_param)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - (-0.81410299), 10e-2)
        self.assertLess(sigma_post_mean - 9.25442675, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)


class SMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        self.observation = [np.array(9.8)]

        # set up for the Bernton et al. implementation:
        # define a distance function
        stat_calc_2 = Identity(degree=1, cross=False)
        self.dist_calc_2 = MMD(stat_calc_2)
        # create fake observed data
        seed = 12
        self.rng = np.random.RandomState(seed)
        self.observation_2 = self.rng.normal(loc=0, scale=1, size=10)
        self.observation_2 = [x for x in self.observation_2]

    def test_sample_delmoral(self):
        # use the SMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape

        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))

        # self.assertEqual((mu_post_mean, sigma_post_mean), (,))

        # use the SMCABC scheme for T = 2
        T, n_sample, n_simulate = 2, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, - 0.8888295384029634, delta=10e-3)
        self.assertAlmostEqual(sigma_post_mean, 4.299346466029422, delta=10e-3)

        self.assertEqual(journal.number_of_simulations[-1], 19)

        # try now with the r-hit kernel version 1:
        T, n_sample, n_simulate = 2, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate, which_mcmc_kernel=1)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -0.6507386970288184, delta=10e-3)
        self.assertAlmostEqual(sigma_post_mean, 6.253446572247367, delta=10e-3)

        self.assertEqual(journal.number_of_simulations[-1], 56)

        # try now with the r-hit kernel version 2:
        T, n_sample, n_simulate = 2, 10, 1
        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate, which_mcmc_kernel=2)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -0.5486451602421536, delta=10e-3)
        self.assertAlmostEqual(sigma_post_mean, 3.633148439032683, delta=10e-3)

        self.assertFalse(journal.number_of_simulations == 0)

    def test_sample_bernton(self):
        # check using the standard MCMC kernel raises error:
        sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="Bernton")
        with self.assertRaises(RuntimeError):
            journal = sampler.sample([self.observation_2], 1, 10, 10, which_mcmc_kernel=0, alpha=0.5,
                                     epsilon_final=0)

        # try now with the r-hit kernel version 1:
        T, n_sample, n_simulate = 3, 10, 10
        sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="Bernton")
        journal = sampler.sample([self.observation_2], T, n_sample, n_simulate, which_mcmc_kernel=1, alpha=0.5,
                                 epsilon_final=0)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -0.7294075767448996, delta=10e-3)
        self.assertAlmostEqual(sigma_post_mean, 3.406347345226374, delta=10e-3)

        self.assertEqual(journal.number_of_simulations[-1], 2286)

        # try now with the r-hit kernel version 2:
        T, n_sample, n_simulate = 3, 10, 10
        sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="Bernton")
        journal = sampler.sample([self.observation_2], T, n_sample, n_simulate, which_mcmc_kernel=2,
                                 epsilon_final=0.1)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertAlmostEqual(mu_post_mean, -2.1412732987491303, delta=10e-3)
        self.assertAlmostEqual(sigma_post_mean, 5.146988585331478, delta=10e-3)

        self.assertEqual(journal.number_of_simulations[-1], 127)

    def test_restart_from_journal_delmoral(self):
        n_sample, n_simulate = 10, 1
        # loop over standard MCMC kernel, r-hit kernel version 1 and r-hit kernel version 2
        for which_mcmc_kernel in [0, 1, 2]:
            # 2 steps with intermediate journal:
            sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
            journal_intermediate = sampler.sample([self.observation], 2, n_sample, n_simulate,
                                                  which_mcmc_kernel=which_mcmc_kernel)
            journal_intermediate.save("tmp.jnl")
            journal_final_1 = sampler.sample([self.observation], 1, n_sample, n_simulate,
                                             which_mcmc_kernel=which_mcmc_kernel,
                                             journal_file="tmp.jnl")

            # 2 steps directly
            sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1)
            journal_final_2 = sampler.sample([self.observation], 3, n_sample, n_simulate, full_output=1,
                                             which_mcmc_kernel=which_mcmc_kernel)
            self.assertEqual(journal_final_1.configuration["epsilon_arr"], journal_final_2.configuration["epsilon_arr"])
            self.assertEqual(journal_final_1.posterior_mean()['mu'], journal_final_2.posterior_mean()['mu'])

    def test_restart_from_journal_bernton(self):
        n_sample, n_simulate = 10, 10
        # loop over standard MCMC kernel, r-hit kernel version 1 and r-hit kernel version 2
        for which_mcmc_kernel in [1, 2]:
            # 2 steps with intermediate journal:
            sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="Bernton")
            journal_intermediate = sampler.sample([self.observation_2], 1, n_sample, n_simulate,
                                                  which_mcmc_kernel=which_mcmc_kernel)
            journal_intermediate.save("tmp.jnl")
            journal_final_1 = sampler.sample([self.observation_2], 1, n_sample, n_simulate,
                                             which_mcmc_kernel=which_mcmc_kernel,
                                             journal_file="tmp.jnl")

            # 2 steps directly
            sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="Bernton")
            journal_final_2 = sampler.sample([self.observation_2], 2, n_sample, n_simulate, full_output=1,
                                             which_mcmc_kernel=which_mcmc_kernel)
            self.assertEqual(journal_final_1.configuration["epsilon_arr"], journal_final_2.configuration["epsilon_arr"])
            self.assertEqual(journal_final_1.posterior_mean()['mu'], journal_final_2.posterior_mean()['mu'])

    def test_errors(self):
        with self.assertRaises(RuntimeError):
            sampler = SMCABC([self.model], [self.dist_calc_2], self.backend, seed=1, version="DelMoral")
        with self.assertRaises(RuntimeError):
            sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1, version="Ciao")

        sampler = SMCABC([self.model], [self.dist_calc], self.backend, seed=1, version="Bernton")
        with self.assertRaises(RuntimeError):
            journal = sampler.sample([self.observation], 1, 10, 10, which_mcmc_kernel=4, alpha=0.5,
                                     epsilon_final=0)


class APMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        # self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

    def test_sample(self):
        # use the APMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = APMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate, alpha=.9)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))

        self.assertFalse(journal.number_of_simulations == 0)

        # use the APMCABC scheme for T = 2
        T, n_sample, n_simulate = 2, 10, 1
        sampler = APMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], T, n_sample, n_simulate, alpha=.9)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - (-3.397848324005792), 10e-2)
        self.assertLess(sigma_post_mean - 6.451434816944525, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)


class RSMCABCTests(unittest.TestCase):
    def setUp(self):
        # find spark and initialize it
        self.backend = BackendDummy()

        # define a uniform prior distribution
        mu = Uniform([[-5.0], [5.0]], name='mu')
        sigma = Uniform([[0.0], [10.0]], name='sigma')
        # define a Gaussian model
        self.model = Normal([mu, sigma])

        # define a distance function
        stat_calc = Identity(degree=2, cross=False)
        self.dist_calc = Euclidean(stat_calc)

        # create fake observed data
        # self.observation = self.model.forward_simulate(1, np.random.RandomState(1))[0].tolist()
        self.observation = [np.array(9.8)]

    def test_sample(self):
        # use the RSMCABC scheme for T = 1
        steps, n_sample, n_simulate = 1, 10, 1
        sampler = RSMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))

        self.assertFalse(journal.number_of_simulations == 0)

        # self.assertEqual((mu_post_mean, sigma_post_mean), (,))

        # use the RSMCABC scheme for T = 2
        steps, n_sample, n_simulate = 2, 10, 1
        sampler = RSMCABC([self.model], [self.dist_calc], self.backend, seed=1)
        journal = sampler.sample([self.observation], steps, n_sample, n_simulate)
        sampler.sample_from_prior(rng=np.random.RandomState(1))
        mu_post_sample, sigma_post_sample, post_weights = np.array(journal.get_parameters()['mu']), np.array(
            journal.get_parameters()['sigma']), np.array(journal.get_weights())

        # Compute posterior mean
        mu_post_mean, sigma_post_mean = journal.posterior_mean()['mu'], journal.posterior_mean()['sigma']

        # test shape of sample
        mu_sample_shape, sigma_sample_shape, weights_sample_shape = (len(mu_post_sample), mu_post_sample[0].shape[1]), \
                                                                    (len(sigma_post_sample),
                                                                     sigma_post_sample[0].shape[1]), post_weights.shape
        self.assertEqual(mu_sample_shape, (10, 1))
        self.assertEqual(sigma_sample_shape, (10, 1))
        self.assertEqual(weights_sample_shape, (10, 1))
        self.assertLess(mu_post_mean - 1.52651600439, 10e-2)
        self.assertLess(sigma_post_mean - 6.49994754262, 10e-2)

        self.assertFalse(journal.number_of_simulations == 0)


if __name__ == '__main__':
    unittest.main()



import torch.nn as nn
import torch.nn.functional as F


class ContrastiveLoss(nn.Module):
    """
    Contrastive loss
    Takes embeddings of two samples and a target label == 1 if samples are from the same class and label == 0 otherwise.

    Code from https://github.com/adambielski/siamese-triplet"""

    def __init__(self, margin):
        super(ContrastiveLoss, self).__init__()
        self.margin = margin
        self.eps = 1e-9

    def forward(self, output1, output2, target, size_average=True):
        distances = (output2 - output1).pow(2).sum(1)  # squared distances
        losses = 0.5 * (target.float() * distances +
                        (1 + -1 * target).float() * F.relu(self.margin - (distances + self.eps).sqrt()).pow(2))
        return losses.mean() if size_average else losses.sum()


class TripletLoss(nn.Module):
    """
    Triplet loss
    Takes embeddings of an anchor sample, a positive sample and a negative sample.

    Code from https://github.com/adambielski/siamese-triplet"""

    def __init__(self, margin):
        super(TripletLoss, self).__init__()
        self.margin = margin

    def forward(self, anchor, positive, negative, size_average=True):
        distance_positive = (anchor - positive).pow(2).sum(1)  # .pow(.5)
        distance_negative = (anchor - negative).pow(2).sum(1)  # .pow(.5)
        losses = F.relu(distance_positive - distance_negative + self.margin)
        return losses.mean() if size_average else losses.sum()



"""Tests here example with MPI."""

import unittest

from abcpy.backends import BackendMPI


def setUpModule():
    '''
    If an exception is raised in a setUpModule then none of
    the tests in the module will be run.

    This is useful because the teams run in a while loop on initialization
    only responding to the scheduler's commands and will never execute anything else.

    On termination of scheduler, the teams call quit() that raises a SystemExit().
    Because of the behaviour of setUpModule, it will not run any unit tests
    for the team and we now only need to write unit-tests from the scheduler's
    point of view.
    '''
    global backend_mpi
    backend_mpi = BackendMPI()


class ExampleGaussianMPITest(unittest.TestCase):
    def test_example(self):
        from examples.backends.mpi.pmcabc_gaussian import infer_parameters
        journal = infer_parameters(backend_mpi, steps=3, n_sample=50)
        test_result = journal.posterior_mean()['mu']
        expected_result = 174.94717012502286
        self.assertAlmostEqual(test_result, expected_result)


if __name__ == '__main__':
    unittest.main()



import logging

import torch
from tqdm import tqdm


def fit(train_loader, model, loss_fn, optimizer, scheduler, n_epochs, cuda, val_loader=None, early_stopping=False,
        epochs_early_stopping_interval=1, start_epoch_early_stopping=10, start_epoch_training=0, use_tqdm=True):
    """
    Basic function to train a neural network given a train_loader, a loss function and an optimizer.

    Loaders, model, loss function and metrics should work together for a given task,
    i.e. The model should be able to process data output of loaders,
    loss function should process target output of loaders and outputs from the model

    Adapted from https://github.com/adambielski/siamese-triplet
    """

    logger = logging.getLogger("NN Trainer")
    if early_stopping and val_loader is not None:
        validation_loss_list = []
    if early_stopping and val_loader is None:
        raise RuntimeError("You cannot perform early stopping if a validation loader is not provided to the training "
                           "routine")

    for epoch in range(0, start_epoch_training):
        scheduler.step()

    for epoch in tqdm(range(start_epoch_training, n_epochs), disable=not use_tqdm):
        # Train stage
        train_loss = train_epoch(train_loader, model, loss_fn, optimizer, cuda)

        logger.debug('Epoch: {}/{}. Train set: Average loss: {:.4f}'.format(epoch + 1, n_epochs, train_loss))

        # Validation stage
        if val_loader is not None:
            val_loss = test_epoch(val_loader, model, loss_fn, cuda)

            logger.debug('Epoch: {}/{}. Validation set: Average loss: {:.4f}'.format(epoch + 1, n_epochs, val_loss))

            # early stopping:
            if early_stopping and (epoch + 1) % epochs_early_stopping_interval == 0:
                validation_loss_list.append(val_loss)  # save the previous validation loss. It is actually
                # we need to have at least two saved test losses for performing early stopping (in which case we know
                # we have saved the previous state_dict as well).
                if epoch + 1 >= start_epoch_early_stopping and len(validation_loss_list) > 1:
                    if validation_loss_list[-1] > validation_loss_list[-2]:
                        logger.info("Training has been early stopped at epoch {}.".format(epoch + 1))
                        # reload the previous state dict:
                        model.load_state_dict(net_state_dict)
                        break  # stop training
                # if we did not stop: update the state dict to the next value
                net_state_dict = model.state_dict()

        scheduler.step()


def train_epoch(train_loader, model, loss_fn, optimizer, cuda):
    """Function implementing the training in one epoch.

    Adapted from https://github.com/adambielski/siamese-triplet
    """
    model.train()
    total_loss = 0

    for batch_idx, (data, target) in enumerate(train_loader):
        target = target if len(target) > 0 else None
        if not type(data) in (tuple, list):
            data = (data,)
        if cuda:
            data = tuple(d.cuda() for d in data)
            if target is not None:
                target = target.cuda()

        optimizer.zero_grad()
        outputs = model(*data)

        if type(outputs) not in (tuple, list):
            outputs = (outputs,)

        loss_inputs = outputs
        if target is not None:
            target = (target,)
            loss_inputs += target

        loss_outputs = loss_fn(*loss_inputs)
        loss = loss_outputs[0] if type(loss_outputs) in (tuple, list) else loss_outputs
        total_loss += loss.item()
        loss.backward()
        optimizer.step()

    return total_loss / (batch_idx + 1)  # divide here by the number of elements in the batch.


def test_epoch(val_loader, model, loss_fn, cuda):
    """Function implementing the computation of the validation error, in batches.

    Adapted from https://github.com/adambielski/siamese-triplet
    """
    with torch.no_grad():
        model.eval()
        val_loss = 0
        for batch_idx, (data, target) in enumerate(val_loader):
            target = target if len(target) > 0 else None
            if not type(data) in (tuple, list):
                data = (data,)
            if cuda:
                data = tuple(d.cuda() for d in data)
                if target is not None:
                    target = target.cuda()

            outputs = model(*data)

            if type(outputs) not in (tuple, list):
                outputs = (outputs,)
            loss_inputs = outputs
            if target is not None:
                target = (target,)
                loss_inputs += target

            loss_outputs = loss_fn(*loss_inputs)
            loss = loss_outputs[0] if type(loss_outputs) in (tuple, list) else loss_outputs
            val_loss += loss.item()

    return val_loss / (batch_idx + 1)  # divide here by the number of elements in the batch.



import unittest

import numpy as np

from abcpy.transformers import DummyTransformer, BoundedVarTransformer


class DummyTransformerTests(unittest.TestCase):
    def test(self):
        transformer = DummyTransformer()
        x = [np.array([3.2]), np.array([2.4])]
        self.assertEqual(x, transformer.transform(x))
        self.assertEqual(x, transformer.inverse_transform(x))
        self.assertEqual(0, transformer.jac_log_det_inverse_transform(x))
        self.assertEqual(0, transformer.jac_log_det(x))


class BoundedVarTransformerTests(unittest.TestCase):
    def setUp(self):
        self.transformer_lower_bounded = BoundedVarTransformer(lower_bound=np.array([0, 0]),
                                                               upper_bound=np.array([None, None]))
        self.transformer_two_sided = BoundedVarTransformer(lower_bound=np.array([0, 0]), upper_bound=np.array([10, 10]))
        self.transformer_mixed = BoundedVarTransformer(lower_bound=np.array([0, 0]), upper_bound=np.array([10, None]))
        self.transformer_dummy = BoundedVarTransformer(lower_bound=np.array([None, None]),
                                                       upper_bound=np.array([None, None]))
        self.list_transformers = [self.transformer_dummy, self.transformer_mixed,
                                  self.transformer_two_sided, self.transformer_lower_bounded]

    def test(self):
        x = [np.array([3.2]), np.array([2.4])]
        for transformer in self.list_transformers:
            self.assertEqual(len(x), len(transformer.inverse_transform(transformer.transform(x))))
            self.assertTrue(np.allclose(np.array(x), np.array(transformer.inverse_transform(transformer.transform(x)))))
            self.assertAlmostEqual(transformer.jac_log_det(x),
                                   transformer.jac_log_det_inverse_transform(transformer.transform(x)), delta=1e-7)

        # test dummy transformer actually does nothing:
        self.assertEqual(x, self.transformer_dummy.transform(x))
        self.assertEqual(x, self.transformer_dummy.inverse_transform(x))
        self.assertEqual(0, self.transformer_dummy.jac_log_det_inverse_transform(x))
        self.assertEqual(0, self.transformer_dummy.jac_log_det(x))

    def test_errors(self):
        with self.assertRaises(RuntimeError):
            transformer = BoundedVarTransformer(lower_bound=[0, 0], upper_bound=[10, 10])
        with self.assertRaises(RuntimeError):
            transformer = BoundedVarTransformer(lower_bound=np.array([0, 0]), upper_bound=np.array([10]))
        with self.assertRaises(NotImplementedError):
            transformer = BoundedVarTransformer(lower_bound=np.array([None, 0]), upper_bound=np.array([10, 10]))
        with self.assertRaises(RuntimeError):
            self.transformer_lower_bounded.transform(x=[np.array([3.2]), np.array([-2.4])])
        with self.assertRaises(RuntimeError):
            self.transformer_two_sided.transform(x=[np.array([13.2]), np.array([2.4])])


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod


class Backend(metaclass=ABCMeta):
    """
    This is the base class for every parallelization backend. It essentially
    resembles the map/reduce API from Spark.

    An idea for the future is to implement a MPI version of the backend with the
    hope to be more complient with standard HPC infrastructure and a potential
    speed-up.

    """

    @abstractmethod
    def parallelize(self, list):
        """
        This method distributes the list on the available workers and returns a
        reference object.

        The list should be split into number of workers many parts. Each
        part should then be sent to a separate worker node.

        Parameters
        ----------
        list: Python list
            the list that should get distributed on the worker nodes
        Returns
        -------
        PDS class (parallel data set)
            A reference object that represents the parallelized list
        """

        raise NotImplementedError

    @abstractmethod
    def broadcast(self, object):
        """
        Send object to all worker nodes without splitting it up.

        Parameters
        ----------
        object: Python object
            An abitrary object that should be available on all workers

        Returns
        -------
        BDS class (broadcast data set)
            A reference to the broadcasted object
        """

        raise NotImplementedError

    @abstractmethod
    def map(self, func, pds):
        """
        A distributed implementation of map that works on parallel data sets (PDS).

        On every element of pds the function func is called.

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDS class
            A parallel data set to which func should be applied
        
        Returns
        -------
        PDS class
            a new parallel data set that contains the result of the map
        """

        raise NotImplementedError

    @abstractmethod
    def collect(self, pds):
        """
        Gather the pds from all the workers, send it to the master and return it as a standard Python list.

        Parameters
        ----------
        pds: PDS class
            a parallel data set
            
        Returns
        -------
        Python list
            all elements of pds as a list
        """

        raise NotImplementedError


class PDS:
    """
    The reference class for parallel data sets (PDS).
    """

    @abstractmethod
    def __init__(self):
        raise NotImplementedError


class BDS:
    """
    The reference class for broadcast data set (BDS).
    """

    @abstractmethod
    def __init__(self):
        raise NotImplementedError

    @abstractmethod
    def value(self):
        """
        This method should return the actual object that the broadcast data set represents. 
        """
        raise NotImplementedError


class BackendDummy(Backend):
    """
    This is a dummy parallelization backend, meaning it doesn't parallelize
    anything. It is mainly implemented for testing purpose.

    """

    def __init__(self):
        pass

    def parallelize(self, python_list):
        """
        This actually does nothing: it just wraps the Python list into dummy pds (PDSDummy).

        Parameters
        ----------
        python_list: Python list
        Returns
        -------        
        PDSDummy (parallel data set)
        """

        return PDSDummy(python_list)

    def broadcast(self, object):
        """
        This actually does nothing: it just wraps the object into BDSDummy.

        Parameters
        ----------
        object: Python object
        
        Returns
        -------        
        BDSDummy class
        """

        return BDSDummy(object)

    def map(self, func, pds):
        """
        This is a wrapper for the Python internal map function.

        Parameters
        ----------
        func: Python func
            A function that can be applied to every element of the pds
        pds: PDSDummy class
            A pseudo-parallel data set to which func should be applied
        
        Returns
        -------
        PDSDummy class
            a new pseudo-parallel data set that contains the result of the map
        """

        result_map = map(func, pds.python_list)
        result_pds = PDSDummy(list(result_map))
        return result_pds

    def collect(self, pds):
        """
        Returns the Python list stored in PDSDummy

        Parameters
        ----------
        pds: PDSDummy class
            a pseudo-parallel data set
        Returns
        -------
        Python list
            all elements of pds as a list
        """

        return pds.python_list


class PDSDummy(PDS):
    """
    This is a wrapper for a Python list to fake parallelization.
    """

    def __init__(self, python_list):
        self.python_list = python_list


class BDSDummy(BDS):
    """
    This is a wrapper for a Python object to fake parallelization.
    """

    def __init__(self, object):
        self.object = object

    def value(self):
        return self.object


class NestedParallelizationController:
    @abstractmethod
    def nested_execution(self):
        raise NotImplementedError

    @abstractmethod
    def run_nested(self, func, *args, **kwargs):
        raise NotImplementedError



"""Tests here the examples which do not require parallelization."""

import unittest
try:
    import torch
except ImportError:
    has_torch = False
else:
    has_torch = True


class ExampleApproxLhdTest(unittest.TestCase):
    def test_pmc(self):
        from examples.approx_lhd.pmc_hierarchical_models import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["school_location"]
        expected_result = 0.2569671145768137
        self.assertAlmostEqual(test_result, expected_result)


class ExampleBackendsTest(unittest.TestCase):
    def test_dummy(self):
        from examples.backends.dummy.pmcabc_gaussian import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["mu"]
        expected_result = 175.00683044068612
        self.assertAlmostEqual(test_result, expected_result)


class ExampleExtensionsModelsTest(unittest.TestCase):
    def test_cpp(self):
        from examples.extensions.models.gaussian_cpp.pmcabc_gaussian_model_simple import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["mu"]
        expected_result = 173.74453347475725
        self.assertAlmostEqual(test_result, expected_result)

    def test_f90(self):
        from examples.extensions.models.gaussian_f90.pmcabc_gaussian_model_simple import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["mu"]
        # note that the f90 example does not always yield the same result on some machines, even if it uses random seed
        expected_result = 173.84265330966315
        self.assertAlmostEqual(test_result, expected_result, delta=3)

    def test_python(self):
        from examples.extensions.models.gaussian_python.pmcabc_gaussian_model_simple import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["mu"]
        expected_result = 175.00683044068612
        self.assertAlmostEqual(test_result, expected_result)

    def test_R(self):
        import os
        print(os.getcwd())
        from examples.extensions.models.gaussian_R.pmcabc_gaussian_model_simple import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["mu"]
        expected_result = 173.4192372459506
        self.assertAlmostEqual(test_result, expected_result)


class ExampleExtensionsPerturbationKernelsTest(unittest.TestCase):
    def test_pmcabc_perturbation_kernel(self):
        from examples.extensions.perturbationkernels.pmcabc_perturbation_kernels import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["schol_without_additional_effects"]
        expected_result = 1.9492397683665226
        self.assertAlmostEqual(test_result, expected_result)


class ExampleHierarchicalModelsTest(unittest.TestCase):
    def test_pmcabc(self):
        from examples.hierarchicalmodels.pmcabc_inference_on_multiple_sets_of_obs import infer_parameters
        journal = infer_parameters(steps=1, n_sample=50)
        test_result = journal.posterior_mean()["schol_without_additional_effects"]
        expected_result = 1.9492397683665226
        self.assertAlmostEqual(test_result, expected_result)


class ExampleModelSelectionTest(unittest.TestCase):
    def test_random_forest(self):
        from examples.modelselection.randomforest_modelselections import infer_model
        model, model_prob = infer_model()
        expected_result = 0.8704000000000001
        # this is not fully reproducible, there are some fluctuations in the estimated value
        self.assertAlmostEqual(model_prob[0], expected_result, delta=0.05)


class ExampleStatisticsLearningTest(unittest.TestCase):
    def test_pmcabc(self):
        if has_torch:
            from examples.statisticslearning.pmcabc_gaussian_statistics_learning import infer_parameters
            journal = infer_parameters(steps=1, n_sample=50)
            test_result = journal.posterior_mean()["mu"]
            expected_result = 172.52136853079725
            self.assertAlmostEqual(test_result, expected_result)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import unittest
import warnings

from abcpy.distances import Euclidean, PenLogReg, LogReg, Wasserstein, SlicedWasserstein, GammaDivergence, MMD, \
    EnergyDistance, KLDivergence, SquaredHellingerDistance
from abcpy.statistics import Identity


class EuclideanTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=1, cross=0)
        self.distancefunc = Euclidean(self.stat_calc)

    def test_distance(self):
        # test simple distance computation
        a = [[0, 0, 0], [0, 0, 0]]
        b = [[0, 0, 0], [0, 0, 0]]
        c = [[1, 1, 1], [1, 1, 1]]
        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, b)
        self.assertRaises(TypeError, self.distancefunc.distance, a, 3.4)

        # test input has different dimensionality
        self.assertRaises(BaseException, self.distancefunc.distance, a, np.array([[0, 0], [1, 2]]))
        self.assertRaises(BaseException, self.distancefunc.distance, a, np.array([[0, 0, 0], [1, 2, 3], [4, 5, 6]]))

        # test whether they compute correct values
        self.assertTrue(self.distancefunc.distance(a, b) == np.array([0]))
        self.assertTrue(self.distancefunc.distance(a, c) == np.array([1.7320508075688772]))

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)


class PenLogRegTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=1, cross=False)
        self.distancefunc = PenLogReg(self.stat_calc)
        self.rng = np.random.RandomState(1)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10
        d3 = 0.5 * self.rng.randn(95, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()
        d3 = d3.tolist()
        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        # completely separable datasets should have a distance of 1.0
        self.assertEqual(self.distancefunc.distance(d1, d2), 1.0)

        # equal data sets should have a distance of 0.0
        self.assertEqual(self.distancefunc.distance(d1, d1), 0.0)

        # equal data sets should have a distance of 0.0; check that in case where n_samples is not a multiple of n_folds
        # in cross validation (10)
        self.assertEqual(self.distancefunc.distance(d3, d3), 0.0)

        # check if it returns the correct error when the number of datasets:
        self.assertRaises(RuntimeError, self.distancefunc.distance, d1, d3)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == 1.0)


class LogRegTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.distancefunc = LogReg(self.stat_calc, seed=1)
        self.rng = np.random.RandomState(1)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        # completely separable datasets should have a distance of 1.0
        self.assertEqual(self.distancefunc.distance(d1, d2), 1.0)

        # equal data sets should have a distance of 0.0
        self.assertEqual(self.distancefunc.distance(d1, d1), 0.0)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == 1.0)


class WassersteinTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.distancefunc = Wasserstein(self.stat_calc)
        self.rng = np.random.RandomState(1)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        self.assertEqual(self.distancefunc.distance(d1, d2), 28.623685155319652)

        # equal data sets should have a distance of approximately 0.0; it won't be exactly 0 due to numerical rounding
        self.assertAlmostEqual(self.distancefunc.distance(d1, d1), 0.0, delta=1e-5)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)


class SlicedWassersteinTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.rng = np.random.RandomState(1)
        self.distancefunc = SlicedWasserstein(self.stat_calc, rng=self.rng)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        self.assertAlmostEqual(self.distancefunc.distance(d1, d2), 12.604402810464576)

        # equal data sets should have a distance of approximately 0.0; it won't be exactly 0 due to numerical rounding
        self.assertAlmostEqual(self.distancefunc.distance(d1, d1), 0.0, delta=1e-5)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)


class GammaDivergenceTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.rng = np.random.RandomState(1)
        # notice: the GammaDivergence estimator becomes infinity when one element in the s1 dataset is equal to one in
        # the s2 dataset and k=1 (as the distance between those two would be 0, which gives infinity when dividing)
        self.distancefunc_k1 = GammaDivergence(self.stat_calc, gam=0.1, k=1)
        self.distancefunc_k2 = GammaDivergence(self.stat_calc, gam=0.1, k=2)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10
        d3 = 0.5 * self.rng.randn(100, 2) + 10
        d3[0] = d1[0]  # set one single element to be equal
        d4 = 0.5 * self.rng.randn(100, 2) + 10
        d4[0] = d4[1] = d1[0]  # set two elements to be equal
        d5 = 0.5 * self.rng.randn(100, 2) + 10
        d5[0] = d5[1] = d5[2] = d1[0]  # set three elements to be equal

        d1 = d1.tolist()
        d2 = d2.tolist()
        d3 = d3.tolist()
        d4 = d4.tolist()
        d5 = d5.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc_k1.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc_k1.distance, d1, 3.4)

        self.assertRaises(ValueError, self.distancefunc_k2.distance, [d1[i] for i in range(2)],
                          [d2[i] for i in range(2)])

        # print(self.distancefunc.distance(d1, d2))
        self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d2), 11.781045117610248)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d2), 9.898652219975974)

        # check identical dataset
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d1), - np.inf)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d3), - np.inf)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d2, d4), + np.inf)
            self.assertAlmostEqual(self.distancefunc_k2.distance(d2, d5), + np.inf)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d1), -1.8398859674447485)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d3), 9.904082614090905)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc_k1.dist_max() == np.inf)


class KLDivergenceTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.rng = np.random.RandomState(1)
        # notice: the KLDivergenceTests estimator becomes infinity when one element in the s1 dataset is equal to one
        # in the s2 dataset and k=1 (as the distance between those two would be 0, which gives infinity when dividing)
        self.distancefunc_k1 = KLDivergence(self.stat_calc, k=1)
        self.distancefunc_k2 = KLDivergence(self.stat_calc, k=2)

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10
        d3 = 0.5 * self.rng.randn(100, 2) + 10
        d3[0] = d1[0]  # set one single element to be equal

        d1 = d1.tolist()
        d2 = d2.tolist()
        d3 = d3.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc_k1.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc_k1.distance, d1, 3.4)

        self.assertRaises(ValueError, self.distancefunc_k2.distance, [d1[i] for i in range(2)],
                          [d2[i] for i in range(2)])

        # print(self.distancefunc.distance(d1, d2))
        self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d2), 11.268656821978732)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d2), 9.632589300292643)

        # check identical dataset
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d1), - np.inf)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d3), - np.inf)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d1), -1.6431603119265736)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d3), 9.629190258363401)

    def test_dist_max(self):
        self.assertTrue(self.distancefunc_k1.dist_max() == np.inf)


class MMDTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.rng = np.random.RandomState(1)
        self.distancefunc = MMD(self.stat_calc)
        self.distancefunc_biased = MMD(self.stat_calc, biased_estimator=True)

    def test_initialization(self):
        self.assertRaises(RuntimeError, MMD, self.stat_calc, 5)
        self.assertRaises(NotImplementedError, MMD, self.stat_calc, "ciao")

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        # print(self.distancefunc.distance(d1, d2))
        self.assertAlmostEqual(self.distancefunc.distance(d1, d2), 0.01078569482366823)
        self.assertAlmostEqual(self.distancefunc_biased.distance(d1, d2), 0.030677837875431546)

        # check identical dataset
        self.assertAlmostEqual(self.distancefunc.distance(d1, d1), - 0.019872124487359623)
        self.assertAlmostEqual(self.distancefunc_biased.distance(d1, d1), 0.0)

        # check if it is symmetric:
        self.assertAlmostEqual(self.distancefunc.distance(d1, d2), self.distancefunc.distance(d2, d1))
        self.assertAlmostEqual(self.distancefunc_biased.distance(d1, d2), self.distancefunc_biased.distance(d2, d1))

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)


class EnergyDistanceTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=2, cross=False)
        self.rng = np.random.RandomState(1)
        self.distancefunc = EnergyDistance(self.stat_calc)
        self.distancefunc_unbiased = EnergyDistance(self.stat_calc, biased_estimator=False)

    def test_initialization(self):
        self.assertRaises(RuntimeError, MMD, self.stat_calc, 5)
        self.assertRaises(NotImplementedError, MMD, self.stat_calc, "ciao")

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10

        d1 = d1.tolist()
        d2 = d2.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc.distance, d1, 3.4)

        # print(self.distancefunc.distance(d1, d2))
        self.assertAlmostEqual(self.distancefunc.distance(d1, d2), 33.76256466781305)
        self.assertAlmostEqual(self.distancefunc_unbiased.distance(d1, d2), 33.41913579594933)

        # check identical dataset
        self.assertAlmostEqual(self.distancefunc.distance(d1, d1), 0.0)
        self.assertAlmostEqual(self.distancefunc_unbiased.distance(d1, d1), -0.31969528899748667)

        # check if it is symmetric:
        self.assertAlmostEqual(self.distancefunc.distance(d1, d2), self.distancefunc.distance(d2, d1))
        self.assertAlmostEqual(self.distancefunc_unbiased.distance(d1, d2),
                               self.distancefunc_unbiased.distance(d2, d1), )

    def test_dist_max(self):
        self.assertTrue(self.distancefunc.dist_max() == np.inf)


class SquaredHellingerDistanceTests(unittest.TestCase):
    def setUp(self):
        self.stat_calc = Identity(degree=1, cross=False)
        self.rng = np.random.RandomState(1)
        # notice: the SquaredHellingerDistance estimator becomes infinity when one element in the s1 dataset is equal to one in
        # the s2 dataset and k=1 (as the distance between those two would be 0, which gives infinity when dividing)
        self.distancefunc_k1 = SquaredHellingerDistance(self.stat_calc, k=1)
        self.distancefunc_k2 = SquaredHellingerDistance(self.stat_calc, k=2)

    # def test_different_distances(self):
    #     d1 = 0.5 * self.rng.randn(100, 2) - 10
    #     d2 = 0.5 * self.rng.randn(100, 2) - 10
    #     d1 = d1.tolist()
    #     for i in range(-20, 40):
    #         d2_list = (d2 + i).tolist()
    #         print(i, self.distancefunc_k1.distance(d1, d2_list))
    #
    #     # the estimator seems to be OK, even if it reaches super small values when the two datasets are from same
    #     # parameters

    def test_distance(self):
        d1 = 0.5 * self.rng.randn(100, 2) - 10
        d2 = 0.5 * self.rng.randn(100, 2) + 10
        d3 = 0.5 * self.rng.randn(100, 2) + 10
        d3[0] = d1[0]  # set one single element to be equal
        d1 = d1.tolist()
        d2 = d2.tolist()
        d3 = d3.tolist()

        # Checks whether wrong input type produces error message
        self.assertRaises(TypeError, self.distancefunc_k1.distance, 3.4, d2)
        self.assertRaises(TypeError, self.distancefunc_k1.distance, d1, 3.4)

        self.assertRaises(ValueError, self.distancefunc_k2.distance, [d1[i] for i in range(2)],
                          [d2[i] for i in range(2)])

        self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d2), 1.123969688201248)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d2), 0.7593257032591699)

        # check infinities when there are same elements in the two datasets
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d1), - np.inf)
            self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d3), - np.inf)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d1), -324.97681257084014)
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d3), -274.4911947796949)

        # check if it is symmetric:
        self.assertAlmostEqual(self.distancefunc_k1.distance(d1, d2), self.distancefunc_k1.distance(d2, d1))
        self.assertAlmostEqual(self.distancefunc_k2.distance(d1, d2), self.distancefunc_k2.distance(d2, d1))

    def test_dist_max(self):
        self.assertTrue(self.distancefunc_k1.dist_max() == 2)


if __name__ == '__main__':
    unittest.main()
